/*  RAxML-VI-HPC (version 2.2) a program for sequential and parallel estimation of phylogenetic trees 
 *  Copyright August 2006 by Alexandros Stamatakis
 *
 *  Partially derived from
 *  fastDNAml, a program for estimation of phylogenetic trees from sequences by Gary J. Olsen
 *  
 *  and 
 *
 *  Programs of the PHYLIP package by Joe Felsenstein.
 *
 *  This program is free software; you may redistribute it and/or modify its
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 * 
 *
 *  For any other enquiries send an Email to Alexandros Stamatakis
 *  Alexandros.Stamatakis@epfl.ch
 *
 *  When publishing work that is based on the results from RAxML-VI-HPC please cite:
 *
 *  Alexandros Stamatakis:"RAxML-VI-HPC: maximum likelihood-based phylogenetic analyses with thousands of taxa and mixed models". 
 *  Bioinformatics 2006; doi: 10.1093/bioinformatics/btl446
 */

#ifndef WIN32
#include <sys/times.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h> 
#endif

#include <limits.h>
#include <math.h>
#include <time.h> 
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "axml.h"

extern int Thorough;
extern infoList iList;

extern char seq_file[1024];
extern char resultFileName[1024];
extern char tree_file[1024];
extern FILE *INFILE;

extern char  workdir[1024];
extern char run_id[128];
extern double masterTime;


#ifdef _USE_PTHREADS
extern volatile int NumberOfThreads;
extern volatile threadEPAdata **epaData;
extern volatile bestBranches **branchData;
extern volatile unsigned char ***globalContiguousTips;
#endif



static double getBranch(tree *tr, double *b, double *bb)
{
  double z = 0.0;

  if(!tr->multiBranch)
    {
      assert(tr->fracchange != -1.0);
      assert(b[0] == bb[0]);
      z = b[0];
      if (z < zmin) 
	z = zmin;      	 
      if(z > zmax)
	z = zmax;
      z = -log(z) * tr->fracchange;
      return z;	
    }
  else
    {       
      int i;
      double x = 0;
      
      for(i = 0; i < tr->numBranches; i++)
	{
	  assert(b[i] == bb[i]);
	  assert(tr->partitionContributions[i] != -1.0);
	  assert(tr->fracchanges[i] != -1.0);
	  x = b[i];
	  if (x < zmin) 
	    x = zmin;      	 
	  if(x > zmax)
	    x = zmax;
	  x = -log(x) * tr->fracchanges[i];
	  z += x * tr->partitionContributions[i];
	}	
      
      return z;
    } 

}


static char *Tree2StringClassifyRec(char *treestr, tree *tr, nodeptr p, int numberOfTipsForInsertion, int *countBranches, 
				    int *inserts, boolean subtreeSummary, boolean *foundTheSubtree, int CUTOFF, boolean originalTree)
{        
  branchInfo *bInf = p->bInf;
  int        i, countQuery = 0;   

  *countBranches = *countBranches + 1;

  if(!subtreeSummary && !originalTree)
    {
      for(i = 0; i < numberOfTipsForInsertion; i++)
	if(bInf->countThem[i] > 0)
	  countQuery++;  
      
      if(countQuery > 0)
	{
	  *treestr++ = '(';   
	  for(i = 0; i < numberOfTipsForInsertion; i++)
	    {
	      if(bInf->countThem[i] > 0)
		{	      
		  sprintf(treestr,"QUERY___%s___%d", tr->nameList[inserts[i]], bInf->countThem[i]);
		  while (*treestr) treestr++;	
		  *treestr++ = ',';
		}
	    }
	}
    }

  if(isTip(p->number, tr->rdta->numsp)) 
    {
      char *nameptr = tr->nameList[p->number];  

      if(subtreeSummary)
	{
	  for(i = 0; i < numberOfTipsForInsertion; i++)
	    {
	      if(p->bInf->executeThem[i] >= CUTOFF)
		foundTheSubtree[i] = TRUE;
	      else
		foundTheSubtree[i] = FALSE;
	    }	  
	}

        
      sprintf(treestr, "%s", nameptr);    
      while (*treestr) treestr++;
    }
  else 
    {                    
      *treestr++ = '(';     
      treestr = Tree2StringClassifyRec(treestr, tr, p->next->back, numberOfTipsForInsertion, 
				       countBranches, inserts,subtreeSummary,foundTheSubtree, CUTOFF, originalTree);     
      *treestr++ = ',';
      treestr = Tree2StringClassifyRec(treestr, tr, p->next->next->back, numberOfTipsForInsertion, 
				       countBranches, inserts, subtreeSummary,foundTheSubtree, CUTOFF, originalTree);          
      *treestr++ = ')';                    

      if(subtreeSummary)
	{  
	  for(i = 0; i < numberOfTipsForInsertion; i++)
	    {
	      if(p->bInf->executeThem[i] >= CUTOFF && 
		 p->next->back->bInf->executeThem[i] < CUTOFF && 
		 p->next->next->back->bInf->executeThem[i] < CUTOFF)
		foundTheSubtree[i] = TRUE;
	      else
		foundTheSubtree[i] = FALSE;
	    }
	}
    }
   
  if(!subtreeSummary && countQuery > 0)
    {
      sprintf(treestr, ":1.0[%s]", p->bInf->branchLabel);
      while (*treestr) treestr++;
      *treestr++ = ')'; 
    }
    
  if(originalTree)
    sprintf(treestr, ":%8.20f[%s", p->bInf->originalBranchLength, p->bInf->branchLabel);
  else
    sprintf(treestr, ":1.0[%s", p->bInf->branchLabel);
  while (*treestr) treestr++;

  assert(!(subtreeSummary == TRUE && originalTree == TRUE));
  assert(!(countQuery > 0 &&  originalTree == TRUE));

  if(subtreeSummary)
    {
      for(i = 0; i < numberOfTipsForInsertion; i++)
	{
	  if(foundTheSubtree[i])
	    {
	      sprintf(treestr," %s", tr->nameList[inserts[i]]);
	      while (*treestr) treestr++;
	    }
	}
    }

  sprintf(treestr, "]");            	 	        
  while (*treestr) treestr++;

  return  treestr;
}


static int findRoot(nodeptr p,  int numberOfTipsForInsertion, int ntips)
{
  if(isTip(p->number, ntips))
    {
      int i;
      for(i = 0; i < numberOfTipsForInsertion; i++)
	if(p->bInf->countThem[i] > 0)
	  return 0;

      return p->number;
    }
  else
    {
      int left;
      assert(p == p->next->next->next);

      left = findRoot(p->next->back, numberOfTipsForInsertion, ntips);

      if(left > 0)
	return left;
      else
	return findRoot(p->next->next->back, numberOfTipsForInsertion, ntips);
    }
}

static void calcSubtree(nodeptr p, int nTips, int numberOfTipsForInsertion)
{
  int i;

  if(isTip(p->number, nTips))    
    {
      for(i = 0; i < numberOfTipsForInsertion; i++)
	p->bInf->executeThem[i] = p->bInf->countThem[i];
    
      return;
    }
  else
    {
      nodeptr q;
      assert(p == p->next->next->next);
      
      q = p->next;

      while(q != p)
	{
	  calcSubtree(q->back, nTips, numberOfTipsForInsertion);	
	  q = q->next;
	}

     
      for(i = 0; i < numberOfTipsForInsertion; i++)
	p->bInf->executeThem[i] = p->next->back->bInf->executeThem[i] + 
	  p->next->next->back->bInf->executeThem[i] + p->bInf->countThem[i];
          
      return;
    }
}


static char *Tree2StringClassify(char *treestr, tree *tr, int numberOfTipsForInsertion, int *inserts, 
				 boolean subtreeSummary, int CUTOFF, int numberOfBranches, branchInfo *bInf, int root, 
				 boolean  originalTree)
{
  nodeptr p;
  int countBranches = 0; 

  if(!subtreeSummary)
    {           
      p = tr->start->back;
      
      assert(!isTip(p->number, tr->mxtips));
      
      *treestr++ = '(';
      treestr = Tree2StringClassifyRec(treestr, tr, p->back, numberOfTipsForInsertion, &countBranches, 
				       inserts, FALSE, (boolean *)NULL, -1, originalTree);
      *treestr++ = ',';
      treestr = Tree2StringClassifyRec(treestr, tr, p->next->back, numberOfTipsForInsertion, &countBranches, 
				       inserts, FALSE, (boolean *)NULL, -1, originalTree);
      *treestr++ = ',';
      treestr = Tree2StringClassifyRec(treestr, tr, p->next->next->back, numberOfTipsForInsertion, &countBranches, 
				       inserts, FALSE, (boolean *)NULL, -1, originalTree);
      *treestr++ = ')';
      *treestr++ = ';';
      
      assert(countBranches == 2 * tr->ntips - 3);
      
      *treestr++ = '\0';
      while (*treestr) treestr++;     
      return  treestr;
    }
  else
    {
      int i, j;
      boolean *foundTheSubtree = (boolean*)malloc(sizeof(boolean) * numberOfTipsForInsertion);

      assert(root > 0);
      assert(originalTree == FALSE);
      
     
      for(i = 0; i < numberOfBranches; i++)
	for(j = 0; j < numberOfTipsForInsertion; j++)
	  bInf[i].executeThem[j] = 0;
           
      p = tr->nodep[root];
	  
      calcSubtree(p, tr->mxtips, numberOfTipsForInsertion);
      calcSubtree(p->back, tr->mxtips, numberOfTipsForInsertion);     	      

      assert(isTip(p->number, tr->mxtips));           
      assert(!isTip(p->back->number, tr->mxtips));
	  
      *treestr++ = '(';
      *treestr++ = '(';	          
      sprintf(treestr, "%s", tr->nameList[p->number]);
      countBranches++;
      while (*treestr) treestr++;
      *treestr++ = ',';
      treestr = Tree2StringClassifyRec(treestr, tr, p->back, numberOfTipsForInsertion, &countBranches, 
				       inserts, TRUE, foundTheSubtree, CUTOFF, originalTree);      
      *treestr++ = ')';
      sprintf(treestr,"[ROOT]");
      while (*treestr) treestr++;
      *treestr++ = ')';
      *treestr++ = ';';
	  
      assert(countBranches == 2 * tr->ntips - 2);
      *treestr++ = '\0';
      while (*treestr) treestr++;
      free(foundTheSubtree);
      return  treestr;   
    }
}




static void markTips(nodeptr p, int *perm, int maxTips)
{
  if(isTip(p->number, maxTips))
    {
      perm[p->number] = 1;
      return;
    }
  else
    {
      nodeptr q = p->next;
      while(q != p)
	{
	  markTips(q->back, perm, maxTips);
	  q = q->next;
	}      
    }
}


static inline double* makeDiagPtable(tree *tr, double z, int dataType, double *rates, double *eign)
{
  double *diagptable = (double*)malloc(sizeof(double) * 4 * tr->partitionData[0].states);  

  calcDiagptable(z, dataType, 4, rates, eign, diagptable);
  
  return diagptable;
}

static inline double scoreGeneric(int states, double *x2, double *x1, double *diagptable, int scale, int dataType)
{
  double  term;    
  int     
    j, 
    k;

  term = 0.0;

 
  for(j = 0; j < 4; j++)
    for(k = 0; k < states; k++)
      term += x1[k] * x2[j * states + k] * diagptable[j * states + k];      
 
  term = log(0.25 * term) + scale * log(minlikelihood);

  return term;
}





#define DIAG 1
#define LEFT 2 




static double mlAlignGeneric(tree *tr, nodeptr p, unsigned char *raw, int unaligned, double *diagptable, boolean writeAlignment, 
			     double currentBest, int nodeNumber, boolean returnAlignment, unsigned char *tempAlignment)
{
  double **scoreMat, path1, path2, result;
  int i, j, nGaps, *refExp;  
  double *al  = &tr->partitionData[0].tipVector[0];
  double *gap;
  double *ref = tr->partitionData[0].xVector[p->number - tr->mxtips -1];
  double *gapVector;  
  char **backtrack = (char **)NULL; 
  int 
    dataType = tr->partitionData[0].dataType,
    tipOffset,
    vectorOffset;

  assert(tr->rateHetModel == GAMMA);

  gap =  &tr->partitionData[0].tipVector[getUndetermined(dataType) * getStates(dataType)];
  tipOffset = tr->partitionData[0].states;
  vectorOffset = tipOffset * 4;

  assert((writeAlignment == FALSE && returnAlignment == FALSE) || (writeAlignment != returnAlignment));

  nGaps = tr->cdta->endsite - unaligned;

  refExp = tr->partitionData[0].expVector[p->number - tr->mxtips - 1];
 
  assert(nGaps > 0);

  scoreMat = (double **)malloc(sizeof(double*) * unaligned);

  for(i = 0; i < unaligned; i++)    
    scoreMat[i] = (double *)calloc(tr->cdta->endsite, sizeof(double)); 

  if(writeAlignment || returnAlignment)    
    {
      backtrack = (char **)malloc(unaligned * sizeof(char*));   
      for(i = 0; i < unaligned; i++)    
	backtrack[i] = (char *)calloc(tr->cdta->endsite, sizeof(char));
    } 

  gapVector = (double *)calloc(nGaps, sizeof(double));

  gapVector[0] = scoreGeneric(tr->partitionData[0].states, &ref[vectorOffset * 0], gap, diagptable, refExp[0], tr->partitionData[0].dataType);
  for(j = 1; j < nGaps; j++)
    gapVector[j] = gapVector[j-1] + scoreGeneric(tr->partitionData[0].states, &ref[vectorOffset * j], gap, diagptable, refExp[j], tr->partitionData[0].dataType);
   
  scoreMat[0][0] =  scoreGeneric(tr->partitionData[0].states, &ref[vectorOffset * 0], &al[tipOffset * raw[0]], diagptable, refExp[j], tr->partitionData[0].dataType);  
  if(writeAlignment || returnAlignment)
    backtrack[0][0] = DIAG;
 
  for(j = 1; j <= nGaps; j++)
    {          
      path1 = gapVector[j-1]   + scoreGeneric(tr->partitionData[0].states, &ref[vectorOffset * j], &al[tipOffset * raw[0]], diagptable, refExp[j], tr->partitionData[0].dataType);
      path2 = scoreMat[0][j-1] + scoreGeneric(tr->partitionData[0].states, &ref[vectorOffset * j], gap, diagptable, refExp[j], tr->partitionData[0].dataType);
      scoreMat[0][j] =  MAX(path1,path2);

      if(writeAlignment || returnAlignment)
	{
	  if(path1 > path2)
	    backtrack[0][j] = DIAG;
	  else
	    backtrack[0][j] = LEFT;
	}
    }          

  for(i = 1; i < unaligned; i++)
    for(j = i; j <= i + nGaps; j++)
      {
	assert(scoreMat[i-1][j-1] != 0);
	if(scoreMat[i][j-1] == 0)
	  {
	    scoreMat[i][j] = scoreMat[i-1][j-1] + scoreGeneric(tr->partitionData[0].states, 
							       &ref[vectorOffset * j], &al[tipOffset * raw[i]], diagptable, refExp[j], tr->partitionData[0].dataType);
	    if(writeAlignment || returnAlignment)
	      backtrack[i][j] = DIAG;
	  }
	else
	  {
	    path1 = scoreMat[i-1][j-1] + scoreGeneric(tr->partitionData[0].states, &ref[vectorOffset * j], &al[tipOffset * raw[i]], diagptable, refExp[j], tr->partitionData[0].dataType);
	    path2 = scoreMat[i][j-1]   + scoreGeneric(tr->partitionData[0].states, &ref[vectorOffset * j], gap,                     diagptable, refExp[j], tr->partitionData[0].dataType);
	    scoreMat[i][j] = MAX(path1, path2);

	    if(writeAlignment || returnAlignment)
	      {
		if(path1 > path2)
		  backtrack[i][j] = DIAG;
		else
		  backtrack[i][j] = LEFT;
	      }	    
	  }	    	
      }
     
  result = scoreMat[unaligned - 1][ tr->cdta->endsite - 1];

  if((writeAlignment && result > currentBest) || returnAlignment)
    {

      unsigned char *tip;      
     

      int counterTip =  tr->cdta->endsite - 1;

      if(writeAlignment)
	tip = tr->yVector[nodeNumber];
      else
	tip = tempAlignment;

      assert(isTip(nodeNumber, tr->mxtips));

      i = unaligned - 1;
      j = tr->cdta->endsite - 1;



      while(i >= 0 && j >= 0)
	{
	  assert(backtrack[i][j] != 0);
	  if(backtrack[i][j] == DIAG)
	    {	     

	      tip[counterTip--]   = raw[i];
	      i--;
	      j--;
	    }
	  else
	    {	      

	      tip[counterTip--]   = 15;
	      j--;
	    }
	}

      while(j >= 0)
	{	 
	  tip[counterTip--]   = 15;	 
	  j--;
	}

      assert(counterTip == -1);      
    }
  
  if(writeAlignment || returnAlignment)
    {
      for(i = 0; i < unaligned; i++)
	free(backtrack[i]);

      free(backtrack);
    }
  
  for(i = 0; i < unaligned; i++)
    free(scoreMat[i]);
  free(scoreMat);
   
  free(gapVector);

  return result;
}





 



static void addInsertion(int branchNumber, double result, int heuristicInsertions, bestBranches *bBr, int j)
{ 
  int i;
  double min;

  assert(heuristicInsertions > 0);

  if(bBr[j].count < heuristicInsertions)
    {
      bBr[j].likelihoods[bBr[j].count] = result;
      bBr[j].branch[bBr[j].count]      = branchNumber;
      bBr[j].count          = bBr[j].count + 1;
      
      if(bBr[j].count == heuristicInsertions)
	{      
	  min = bBr[j].likelihoods[0];
	  bBr[j].minLH = 0;

	  for(i = 1; i < heuristicInsertions; i++)
	    {
	      if(bBr[j].likelihoods[i] < min)
		{
		  min = bBr[j].likelihoods[i];
		  bBr[j].minLH = i;
		}
	    }	  
	}

      return;
    }

  if(result < bBr[j].likelihoods[bBr[j].minLH])
    return;
  
  bBr[j].likelihoods[bBr[j].minLH] = result;
  bBr[j].branch[bBr[j].minLH]      = branchNumber;
          
	 
  min = bBr[j].likelihoods[0];
  bBr[j].minLH = 0;
  for(i = 1; i < heuristicInsertions; i++)
    {
      if(bBr[j].likelihoods[i] < min)
	{
	  min = bBr[j].likelihoods[i];
	  bBr[j].minLH = i;
	}
    }

  return; 
}



static void testInsertThorough(tree *tr, nodeptr r, nodeptr q, int *inserts,  int numberOfTipsForInsertion,
			       double *insertLikelihoods, double *branches, nodeptr *insertNodes, boolean align, double *diagptable, 
			       boolean writeAlignment, unalignedSequences *rawSeqs, int dataType, bestBranches *bBr, boolean bootstrap)
{
  double 
    result,           
    qz[NUM_BRANCHES],
    zqs[NUM_BRANCHES], 
    zrs[NUM_BRANCHES], 
    lzqr, 
    lzqs, 
    lzrs, 
    lzsum, 
    lzq, 
    lzr, 
    lzs, 
    lzmax, 
    zz,      
    defaultArray[NUM_BRANCHES],
    e1[NUM_BRANCHES], 
    e2[NUM_BRANCHES], 
    e3[NUM_BRANCHES];
  
  nodeptr  x;      
  
  int 
    i, 
    j; 

  assert(!tr->grouped);

  x = q->back;  

  for(i = 0; i < tr->numBranches; i++)    
    {
      qz[i] = q->z[i];
      defaultArray[i] = defaultz;
    }
  
  for(j = 0; j < numberOfTipsForInsertion; j++)
    {                
      if(tr->heuristicInsertions >= 0 || (tr->heuristicInsertions == -1 && q->bInf->executeThem[j] == 1))
	{	 	  
	  if(align)
	    {
	      double 
		tempz[NUM_BRANCHES],
		*diagptableTwo = (double *)NULL;
	      unsigned char *tempAlignment  = (unsigned char *)malloc(sizeof(unsigned char) * tr->cdta->endsite);
	      unsigned char *sequenceBuffer = (unsigned char *)malloc(sizeof(unsigned char) * tr->cdta->endsite);
	      
	      nodeptr s = tr->nodep[inserts[j]];	     	     

	      for(i = 0; i < tr->numBranches; i++)
		{		      
		  tempz[i] = sqrt(qz[i]);      
		  
		  if(tempz[i] < zmin) 
		    tempz[i] = zmin;
		  if(tempz[i] > zmax)
		    tempz[i] = zmax;
		}   
	      
	      hookup(r->next,       q, tempz, tr->numBranches);
	      hookup(r->next->next, x, tempz, tr->numBranches);	
	      newviewGeneric(tr, r);
	      
	      mlAlignGeneric(tr, r, rawSeqs[j].seq, rawSeqs[j].length, diagptable, FALSE, insertLikelihoods[j],
			     inserts[j], TRUE, tempAlignment);	     	     
	      
	      assert(s->number == inserts[j]);
	      
	      memcpy(sequenceBuffer,          tr->yVector[inserts[j]], sizeof(char) * tr->cdta->endsite);
	      memcpy(tr->yVector[inserts[j]], tempAlignment,           sizeof(char) * tr->cdta->endsite);	 
	      
	      makenewzGeneric(tr, q, s, defaultArray, iterations, zqs, FALSE);                  
	      makenewzGeneric(tr, x, s, defaultArray, iterations, zrs, FALSE);
	      
	      for(i = 0; i < tr->numBranches; i++)
		{
		  lzqr = (qz[i] > zmin) ? log(qz[i]) : log(zmin); 
		  lzqs = (zqs[i] > zmin) ? log(zqs[i]) : log(zmin);
		  lzrs = (zrs[i] > zmin) ? log(zrs[i]) : log(zmin);
		  lzsum = 0.5 * (lzqr + lzqs + lzrs);
		  
		  lzq = lzsum - lzrs;
		  lzr = lzsum - lzqs;
		  lzs = lzsum - lzqr;
		  lzmax = log(zmax);
		  
		  if      (lzq > lzmax) {lzq = lzmax; lzr = lzqr; lzs = lzqs;} 
		  else if (lzr > lzmax) {lzr = lzmax; lzq = lzqr; lzs = lzrs;}
		  else if (lzs > lzmax) {lzs = lzmax; lzq = lzqs; lzr = lzrs;}          
		  
		  e1[i] = exp(lzq);
		  e2[i] = exp(lzr);
		  e3[i] = exp(lzs);
		}	   	 
	      
	      hookup(r->next,       q, e1, tr->numBranches);
	      hookup(r->next->next, x, e2, tr->numBranches);
	      hookup(r,             s, e3, tr->numBranches); 
	      
	      newviewGeneric(tr, r);	   	  
	      localSmooth(tr, r, smoothings);

	      

	      result = evaluateGeneric(tr, r);
	      
	      /*printf("BR-OPT \t\t %d %f\n", *count, result);*/
	      
	      memcpy(tr->yVector[inserts[j]], sequenceBuffer, sizeof(char) * tr->cdta->endsite);
	      
	      newviewGeneric(tr, r);
	      	     
	      zz = r->z[0];
	     
	      diagptableTwo =  makeDiagPtable(tr, zz, dataType, tr->partitionData[0].gammaRates,tr->partitionData[0].EIGN);	 
	      result = mlAlignGeneric(tr, r, rawSeqs[j].seq, rawSeqs[j].length, diagptable, writeAlignment, insertLikelihoods[j],
				      inserts[j], FALSE, (unsigned char*)NULL);
	     	     	       	      

	      if(result > insertLikelihoods[j])
		{	     
		  insertLikelihoods[j] = result;
		  insertNodes[j]       = q;		 	     		 
		}      	   	    	    
	      	      
	      
	

	      free(diagptableTwo);	      
	      free(tempAlignment);
	      free(sequenceBuffer);
	    }
	  else
	    {    
	      nodeptr s = tr->nodep[inserts[j]];	      	     
	     
	      makenewzGeneric(tr, q, s, defaultArray, iterations, zqs, FALSE);                  
	      
	      
	     
	      makenewzGeneric(tr, x, s, defaultArray, iterations, zrs, FALSE);
	     
	      
	      for(i = 0; i < tr->numBranches; i++)
		{
		  lzqr = (qz[i] > zmin) ? log(qz[i]) : log(zmin);		  
		  lzqs = (zqs[i] > zmin) ? log(zqs[i]) : log(zmin);
		  lzrs = (zrs[i] > zmin) ? log(zrs[i]) : log(zmin);
		  lzsum = 0.5 * (lzqr + lzqs + lzrs);
		  
		  lzq = lzsum - lzrs;
		  lzr = lzsum - lzqs;
		  lzs = lzsum - lzqr;
		  lzmax = log(zmax);
		  
		  if      (lzq > lzmax) {lzq = lzmax; lzr = lzqr; lzs = lzqs;} 
		  else if (lzr > lzmax) {lzr = lzmax; lzq = lzqr; lzs = lzrs;}
		  else if (lzs > lzmax) {lzs = lzmax; lzq = lzqs; lzr = lzrs;}          
		  
		  e1[i] = exp(lzq);
		  e2[i] = exp(lzr);
		  e3[i] = exp(lzs);
		}
	      
	      hookup(r->next,       q, e1, tr->numBranches);
	      hookup(r->next->next, x, e2, tr->numBranches);
	      hookup(r,             s, e3, tr->numBranches);      		     
	     
	      newviewGeneric(tr, r);	     
	      
	      localSmooth(tr, r, smoothings);
	     	      
	      result = evaluateGeneric(tr, r);	 	       
	      
	      if(result > insertLikelihoods[j])
		{	     
		  insertLikelihoods[j] = result;
		  insertNodes[j]       = q;
		  if(bootstrap)
		    {
		      assert(branches);
		      branches[j] = getBranch(tr, r->z, r->back->z);
		    }
		} 
	      
	      if(tr->heuristicInsertions > 0)				
		addInsertion(q->bInf->branchNumber, result, tr->heuristicInsertions, bBr, j);
	       
	    }	 
	}
    }

  hookup(q, x, qz, tr->numBranches);
   
  r->next->next->back = r->next->back = (nodeptr) NULL; 
}



static void testInsertFast(tree *tr, nodeptr r, nodeptr q, int *inserts,  int numberOfTipsForInsertion, 
			   double *insertLikelihoods, nodeptr *insertNodes, boolean align, double *diagptable, 
			   boolean writeAlignment, unalignedSequences *rawSeqs, bestBranches *bBr)
{
  double result;           
  double  qz[NUM_BRANCHES], z[NUM_BRANCHES];
  nodeptr  x;       
  int i;
  x = q->back;     	       
  
  

  assert(!tr->grouped);                    
  
  for(i = 0; i < tr->numBranches; i++)
    {	
      qz[i] = q->z[i];
      z[i] = sqrt(q->z[i]);      
      
      if(z[i] < zmin) 
	z[i] = zmin;
      if(z[i] > zmax)
	z[i] = zmax;
    }        
  
  hookup(r->next,       q, z, tr->numBranches);
  hookup(r->next->next, x, z, tr->numBranches);	                         
  
  newviewGeneric(tr, r);   
  
  for(i = 0; i < numberOfTipsForInsertion; i++)
    {
      if(tr->heuristicInsertions >= 0 || (tr->heuristicInsertions == -1 && q->bInf->executeThem[i] == 1))
	{
	  if(align)	
	    {	     	      
	      result = mlAlignGeneric(tr, r, rawSeqs[i].seq, rawSeqs[i].length, diagptable, writeAlignment, insertLikelihoods[i],
				      inserts[i], FALSE, (unsigned char*)NULL);	   		     
	    }
	  else
	    {	 
	      hookupDefault(r, tr->nodep[inserts[i]], tr->numBranches);
	      result = evaluateGeneric (tr, r);	     
	      
	      r->back = (nodeptr) NULL;
	      tr->nodep[inserts[i]]->back = (nodeptr) NULL;
	    }	 
 
	  if(result > insertLikelihoods[i])
	    {	     
	      insertLikelihoods[i] = result;
	      insertNodes[i]       = q;	     
	    } 
	  
	  if(tr->heuristicInsertions > 0)
	    addInsertion(q->bInf->branchNumber, result, tr->heuristicInsertions, bBr, i);    
	}    
    }
 

  hookup(q, x, qz, tr->numBranches);
  
  r->next->next->back = r->next->back = (nodeptr) NULL;
}

static void addTraverseRob(tree *tr, nodeptr r, nodeptr q, int *inserts,  int numberOfTipsForInsertion,
			   double *insertLikelihoods, double *branches, nodeptr *insertNodes, boolean align, double *diagptable, 
			   boolean writeAlignment, unalignedSequences *rawSeqs, boolean thorough, int dataType, bestBranches *bBr,
			   boolean bootstrap)
{       
  if(thorough)
    testInsertThorough(tr, r, q, inserts,  numberOfTipsForInsertion, insertLikelihoods, branches, insertNodes, align,
		       diagptable, writeAlignment, rawSeqs, dataType, bBr, bootstrap);
  else
    testInsertFast(tr, r, q, inserts,  numberOfTipsForInsertion, insertLikelihoods, insertNodes, align,
		   diagptable, writeAlignment, rawSeqs, bBr);

  if(!isTip(q->number, tr->rdta->numsp))
    {   
      nodeptr a = q->next;

      while(a != q)
	{
	  addTraverseRob(tr, r, a->back,       inserts,  numberOfTipsForInsertion, insertLikelihoods, branches,
			 insertNodes, align, diagptable, writeAlignment, rawSeqs, thorough, dataType, bBr, 
			 bootstrap);
	  a = a->next;
	}      
    }
} 

#ifdef _USE_PTHREADS



static size_t getContiguousVectorLength(tree *tr)
{
  size_t length = 0;
  int model;

  for(model = 0; model < tr->NumberOfModels; model++)
    {     
      size_t realWidth = tr->partitionData[model].upper - tr->partitionData[model].lower;
      int states = tr->partitionData[model].states;

      length += (realWidth * (size_t)states * (size_t)(tr->discreteRateCategories));

      	
    }

  return length;
}
static size_t getContiguousScalingLength(tree *tr)
{
  size_t length = 0;
  int model;

  for(model = 0; model < tr->NumberOfModels; model++)    
    length += tr->partitionData[model].upper - tr->partitionData[model].lower;

  return length;
}



static void gatherLikelihoodVectors(tree *tr, int *countBranches, nodeptr p)
{ 
  if(isTip(p->number, tr->mxtips))    
    {                
      int i;
      
      contiguousBranch *b = tr->GLOBAL_bInfo;
      
      b->branchNumber = *countBranches;	
      
      b->leftNodeNumber = p->number;
      b->rightNodeNumber = p->back->number;
      
      
      if(!p->back->x)
	newviewGeneric(tr, p->back);	
      
      for(i = 0; i < tr->numBranches; i++)
	b->branchLengths[i] = p->z[i];
      
      masterBarrier(THREAD_GATHER_LIKELIHOOD, tr);
      masterBarrier(THREAD_WHO_WANTS_THE_BRANCH, tr);
      
      *countBranches =  *countBranches + 1;
      return;
    }
  else
    {
      nodeptr q;                    
      int i;
      contiguousBranch *b = tr->GLOBAL_bInfo;

      b->branchNumber = *countBranches;

      b->leftNodeNumber = p->number;
      b->rightNodeNumber = p->back->number;	
	            
		 
      if(!p->x)
	newviewGeneric(tr, p);            
	 
      if(!isTip(p->back->number, tr->mxtips))
	{
	  if(!p->back->x)
	    newviewGeneric(tr, p->back);	 
	}
	
      for(i = 0; i < tr->numBranches; i++)
	b->branchLengths[i] = p->z[i];	

      masterBarrier(THREAD_GATHER_LIKELIHOOD, tr);
      masterBarrier(THREAD_WHO_WANTS_THE_BRANCH, tr);
            
      *countBranches =  *countBranches + 1;
           
      q = p->next;

      while(q != p)
	{
	  gatherLikelihoodVectors(tr, countBranches, q->back);	
	  q = q->next;
	}
     
      return;
    }
}
 
static void collectLikelihoodVectors(tree *tr, nodeptr p, int numberOfBranches)
{
  int countBranches = 0;

  gatherLikelihoodVectors(tr, &countBranches, p);

  assert(countBranches == numberOfBranches);
}


static void updateClassify(tree *tr, double *z, boolean *partitionSmoothed, boolean *partitionConverged, double *x1, double *x2, unsigned char *tipX1, unsigned char *tipX2, int tipCase)
{   
  int i;

  boolean smoothedPartitions[NUM_BRANCHES];

  double 
    newz[NUM_BRANCHES], 
    z0[NUM_BRANCHES];
  
  for(i = 0; i < tr->numBranches; i++)   
    z0[i] = z[i];          

  makenewzClassify(tr, newzpercycle, newz, z0, x1, x2, tipX1, tipX2, tipCase, partitionConverged); 

  for(i = 0; i < tr->numBranches; i++)    
    smoothedPartitions[i]  = partitionSmoothed[i];
 
  for(i = 0; i < tr->numBranches; i++)
    {         
      if(!partitionConverged[i])
	{		    
	  if(ABS(newz[i] - z0[i]) > deltaz)  	     
	    smoothedPartitions[i] = FALSE;       
	    	             
	  z[i] = newz[i];	 
	}
    }


  for(i = 0; i < tr->numBranches; i++)    
    partitionSmoothed[i]  = smoothedPartitions[i];  
}


static double localSmoothClassify (tree *tr, int maxtimes, int leftNodeNumber, int rightNodeNumber, int insertionNodeNumber, double *e1, double *e2, double *e3, 
				   contiguousBranch *b)
{ 
  int tipCase;
  
  boolean 
    allSmoothed,
    partitionSmoothed[NUM_BRANCHES],
    partitionConverged[NUM_BRANCHES];

  double 
    result,
    *x1 = (double*)NULL,
    *x2 = (double*)NULL,
    *x3 = (double*)NULL;
	  
  int
    i,
    *ex1 = (int*)NULL,
    *ex2 = (int*)NULL,
    *ex3 = (int*)NULL; 

  unsigned char 
    *tipX1 = (unsigned char *)NULL,
    *tipX2 = (unsigned char *)NULL;
  
  x3  = tr->temporaryVector;
  ex3 = tr->temporaryScaling;
    
  
  for(i = 0; i < tr->numBranches; i++)	
    partitionConverged[i] = FALSE;	

  while (--maxtimes >= 0) 
    {     
      for(i = 0; i < tr->numBranches; i++)	
	partitionSmoothed[i] = TRUE;     
	 	
      /* e3 */
      if(isTip(leftNodeNumber, tr->mxtips) && isTip(rightNodeNumber, tr->mxtips))
	{
	  tipCase = TIP_TIP;
	  
	  tipX1 = tr->contiguousTips[leftNodeNumber];
	  tipX2 = tr->contiguousTips[rightNodeNumber];

	  newviewClassifySpecial(tr, x1, x2, x3, ex1, ex2, ex3, tipX1, tipX2, tipCase, e1, e2);
	}
      else
	{
	  if (isTip(leftNodeNumber, tr->mxtips))
	    {
	      tipCase = TIP_INNER;

	      tipX1 = tr->contiguousTips[leftNodeNumber];
	      
	      x2  = b->right;	     
	      ex2 = b->rightScaling; 	  
	      newviewClassifySpecial(tr, x1, x2, x3, ex1, ex2, ex3, tipX1, tipX2, tipCase, e1, e2);
	    }
	  else 
	    {
	      if(isTip(rightNodeNumber, tr->mxtips))
		{
		  tipCase = TIP_INNER;

		  tipX1 = tr->contiguousTips[rightNodeNumber];
	  
		  x2  = b->left;	 
		  ex2 = b->leftScaling; 
		  newviewClassifySpecial(tr, x1, x2, x3, ex1, ex2, ex3, tipX1, tipX2, tipCase, e2, e1);
		}       
	      else
		{
		  tipCase = INNER_INNER;
		  
		  x1  = b->left;
		  ex1 = b->leftScaling;
		  
		  x2  = b->right;
		  ex2 = b->rightScaling;
		  newviewClassifySpecial(tr, x1, x2, x3, ex1, ex2, ex3, tipX1, tipX2, tipCase, e1, e2);
		}
	    }
	}
	
     

      tipCase = TIP_INNER;
      
      x2    = tr->temporaryVector;
      tipX1 = tr->contiguousTips[insertionNodeNumber];

      updateClassify(tr, e3, partitionSmoothed, partitionConverged, x1, x2, tipX1, tipX2, tipCase);

      /* e1 **********************************************************/

      tipX1 = tr->contiguousTips[insertionNodeNumber];

      if(isTip(rightNodeNumber, tr->mxtips))
	{
	  tipCase = TIP_TIP;

	  tipX2 = tr->contiguousTips[rightNodeNumber];
	}
      else
	{	 
	  tipCase = TIP_INNER;
		  
	  x2  = b->right;
	  ex2 = b->rightScaling;		  	
	}
	
      newviewClassifySpecial(tr, x1, x2, x3, ex1, ex2, ex3, tipX1, tipX2, tipCase, e3, e2);

      if(isTip(leftNodeNumber, tr->mxtips))
	{
	  tipCase = TIP_INNER;

	  tipX1 = tr->contiguousTips[leftNodeNumber];
	}
      else
	{
	  tipCase = INNER_INNER;

	  x1      =  b->left;
	}

      updateClassify(tr, e1, partitionSmoothed, partitionConverged, x1, x3, tipX1, (unsigned char*)NULL, tipCase);

      /* e2 *********************************************************/

      tipX1 = tr->contiguousTips[insertionNodeNumber];

      if(isTip(leftNodeNumber, tr->mxtips))
	{
	  tipCase = TIP_TIP;
	  
	  tipX2 = tr->contiguousTips[leftNodeNumber];
	}
      else
	{	 
	  tipCase = TIP_INNER;
		  
	  x2  = b->left;
	  ex2 = b->leftScaling;		  	
	}
	
      newviewClassifySpecial(tr, x1, x2, x3, ex1, ex2, ex3, tipX1, tipX2, tipCase, e3, e1);

      if(isTip(rightNodeNumber, tr->mxtips))
	{
	  tipCase = TIP_INNER;

	  tipX1 = tr->contiguousTips[rightNodeNumber];
	}
      else
	{
	  tipCase = INNER_INNER;

	  x1      =  b->right;
	}

      updateClassify(tr, e2, partitionSmoothed, partitionConverged, x1, x3, tipX1, (unsigned char*)NULL, tipCase);


      allSmoothed = TRUE;
      for(i = 0; i < tr->numBranches; i++)
	{
	  if(partitionSmoothed[i] == FALSE)
	    allSmoothed = FALSE;
	  else
	    partitionConverged[i] = TRUE;
	}
      
      if(allSmoothed)
	break;
    }

  
  if(isTip(leftNodeNumber, tr->mxtips) && isTip(rightNodeNumber, tr->mxtips))
    {
      tipCase = TIP_TIP;

      tipX1 = tr->contiguousTips[leftNodeNumber];
      tipX2 = tr->contiguousTips[rightNodeNumber];

      newviewClassifySpecial(tr, x1, x2, x3, ex1, ex2, ex3, tipX1, tipX2, tipCase, e1, e2);
    }
  else
    {
      if (isTip(leftNodeNumber, tr->mxtips))
	{
	  tipCase = TIP_INNER;

	  tipX1 = tr->contiguousTips[leftNodeNumber];	  

	  x2  = b->right;	     
	  ex2 = b->rightScaling; 	  
	  newviewClassifySpecial(tr, x1, x2, x3, ex1, ex2, ex3, tipX1, tipX2, tipCase, e1, e2);
	}
      else 
	{
	  if(isTip(rightNodeNumber, tr->mxtips))
	    {
	      tipCase = TIP_INNER;

	      tipX1 = tr->contiguousTips[rightNodeNumber];
	      
	      x2  = b->left;	 
	      ex2 = b->leftScaling; 
	      newviewClassifySpecial(tr, x1, x2, x3, ex1, ex2, ex3, tipX1, tipX2, tipCase, e2, e1);
	    }       
	  else
	    {
	      tipCase = INNER_INNER;
	      
	      x1  = b->left;
	      ex1 = b->leftScaling;
	      
	      x2  = b->right;
	      ex2 = b->rightScaling;
	      newviewClassifySpecial(tr, x1, x2, x3, ex1, ex2, ex3, tipX1, tipX2, tipCase, e1, e2);
	    }
	}
    }
  
     
  tipCase = TIP_INNER;
  
  x2    = tr->temporaryVector;
  tipX1 = tr->contiguousTips[insertionNodeNumber];    
 
  result = evalCL(tr, x2, ex3, tipX1, e3);

  return result;
}

void testInsertThoroughIterative(tree *tr)
{
  int 
    i;

  double 
    defaultArray[NUM_BRANCHES];
  
  boolean 
    executeAll[NUM_BRANCHES];

  for(i = 0; i < tr->branchCounter; i++)
    {
      defaultArray[i] = defaultz;
      executeAll[i] = TRUE;
    }
   
  for(i = 0; i < tr->branchCounter; i++)   	         
    {
      double	    
	result, 
	lzqr, 
	lzqs, 
	lzrs, 
	lzsum, 
	lzq, 
	lzr, 
	lzs, 
	lzmax, 
	zqs[NUM_BRANCHES],
	zrs[NUM_BRANCHES],
	*qz,
	e1[NUM_BRANCHES], 
	e2[NUM_BRANCHES], 
	e3[NUM_BRANCHES];  
	
      int
	tipCase,	   
	insertions,
	leftNodeNumber,
	rightNodeNumber;
	  

      contiguousBranch *b = &(tr->LOCAL_bInfo[i]);
	  
      leftNodeNumber  = b->leftNodeNumber;
      rightNodeNumber = b->rightNodeNumber;
	  
      for(insertions = 0; insertions < tr->numberOfTipsForInsertion; insertions++)
	{ 
	  if(tr->heuristicInsertions >= 0 || (tr->heuristicInsertions == -1 && b->executeThem[insertions] == 1))
	    {	     
	      double 
		*x1 = (double*)NULL,
		*x2 = (double*)NULL,
		*x3 = (double*)NULL;
	      
	      int
		model = 0,
		*ex1 = (int*)NULL,
		*ex2 = (int*)NULL,
		*ex3 = (int*)NULL; 
	      
	      unsigned char 
		*tipX1 = (unsigned char *)NULL,
		*tipX2 = (unsigned char *)NULL;	  	  
	      
	      qz = b->branchLengths;      
	      
	      
	      
	      tipX1 = tr->contiguousTips[tr->inserts[insertions]];
	      
	      if(isTip(leftNodeNumber, tr->mxtips))
		{
		  tipX2 = tr->contiguousTips[leftNodeNumber];
		  x2 = (double*)NULL;
		  tipCase = TIP_TIP;
		}
	      else
		{
		  tipX2 = (unsigned char*)NULL;
		  x2 = b->left;
		  tipCase = TIP_INNER;
		}
	      
	      makenewzClassify(tr, iterations, zqs, defaultArray, (double*)NULL, x2, tipX1, tipX2, tipCase, executeAll);  
	      
	      
	      if(isTip(rightNodeNumber, tr->mxtips))
		{ 
		  x2 = (double*)NULL;
		  
		  tipX2 = tr->contiguousTips[rightNodeNumber];
		  
		  tipCase = TIP_TIP;
		}
	      else
		{
		  tipX2 = (unsigned char*)NULL;
		  x2 = b->right;
		  tipCase = TIP_INNER;
		}
	      
	      makenewzClassify(tr, iterations, zrs, defaultArray, (double*)NULL, x2, tipX1, tipX2, tipCase, executeAll);      
	      
	      
	      for(model = 0; model < tr->numBranches; model++)
		{
		  lzqr = (qz[model] > zmin) ? log(qz[model]) : log(zmin);		  
		  lzqs = (zqs[model] > zmin) ? log(zqs[model]) : log(zmin);
		  lzrs = (zrs[model] > zmin) ? log(zrs[model]) : log(zmin);
		  lzsum = 0.5 * (lzqr + lzqs + lzrs);
		  
		  lzq = lzsum - lzrs;
		  lzr = lzsum - lzqs;
		  lzs = lzsum - lzqr;
		  lzmax = log(zmax);
		  
		  if      (lzq > lzmax) {lzq = lzmax; lzr = lzqr; lzs = lzqs;} 
		  else if (lzr > lzmax) {lzr = lzmax; lzq = lzqr; lzs = lzrs;}
		  else if (lzs > lzmax) {lzs = lzmax; lzq = lzqs; lzr = lzrs;}          
		  
		  e1[model] = exp(lzq);
		  e2[model] = exp(lzr);
		  e3[model] = exp(lzs);
		}
	      
	      x3  = tr->temporaryVector;
	      ex3 = tr->temporaryScaling;
	      
	      if(isTip(leftNodeNumber, tr->mxtips) && isTip(rightNodeNumber, tr->mxtips))
		{
		  tipCase = TIP_TIP;
		  
		  tipX1 = tr->contiguousTips[leftNodeNumber];
		  tipX2 = tr->contiguousTips[rightNodeNumber];
		  
		  newviewClassifySpecial(tr, x1, x2, x3, ex1, ex2, ex3, tipX1, tipX2, tipCase, e1, e2);	
		}
	      else
		{
		  if (isTip(leftNodeNumber, tr->mxtips))
		    {
		      tipCase = TIP_INNER;
		      
		      tipX1 = tr->contiguousTips[leftNodeNumber];
		      
		      x2  = b->right;	     
		      ex2 = b->rightScaling; 
		      newviewClassifySpecial(tr, x1, x2, x3, ex1, ex2, ex3, tipX1, tipX2, tipCase, e1, e2);	
		    }
		  else 
		    {
		      if(isTip(rightNodeNumber, tr->mxtips))
			{
			  tipCase = TIP_INNER;
			  
			  tipX1 = tr->contiguousTips[rightNodeNumber];
			  
			  x2  = b->left;	 
			  ex2 = b->leftScaling;
			  newviewClassifySpecial(tr, x1, x2, x3, ex1, ex2, ex3, tipX1, tipX2, tipCase, e2, e1);	
			}       
		      else
			{
			  tipCase = INNER_INNER;
			  
			  x1  = b->left;
			  ex1 = b->leftScaling;
			  
			  x2  = b->right;
			  ex2 = b->rightScaling;
			  newviewClassifySpecial(tr, x1, x2, x3, ex1, ex2, ex3, tipX1, tipX2, tipCase, e1, e2);	
			}
		    }	      
		}
	      	    
	      result = localSmoothClassify(tr, smoothings, leftNodeNumber, rightNodeNumber, tr->inserts[insertions], e1, e2, e3, b);
	      
	      if(result > tr->epaData->scoreVector[insertions])
		{	     
		  tr->epaData->scoreVector[insertions]  = result;
		  tr->epaData->branchVector[insertions] = b->branchNumber;
		  tr->epaData->insertionBranchLengths[insertions] = getBranch(tr, e3, e3);
		}   	 
	      
	      if(tr->heuristicInsertions > 0)
		addInsertion(b->branchNumber, result, tr->heuristicInsertions, tr->bBr, insertions);
	      
	    }	  
	}
    }
}






static double mlAlignGenericContiguous(tree *tr, double *ref, int *refExp, unsigned char *raw, int unaligned, double *diagptable,
				       double currentBest, unsigned char *tempAlignment)
{
  double **scoreMat, path1, path2, result;
  int i, j, nGaps;  
  double *al  = &tr->partitionData[0].tipVector[0];
  double *gap;  
  double *gapVector;  
  char **backtrack = (char **)NULL; 
  int 
    dataType = tr->partitionData[0].dataType,
    tipOffset,
    vectorOffset;

  assert(tr->rateHetModel == GAMMA);

  gap =  &tr->partitionData[0].tipVector[getUndetermined(dataType) * getStates(dataType)];
  tipOffset = tr->partitionData[0].states;
  vectorOffset = tipOffset * 4;  
 
  nGaps = tr->contiguousScalingLength - unaligned;
    
  assert(nGaps > 0);

  scoreMat = (double **)malloc(sizeof(double*) * unaligned);

  for(i = 0; i < unaligned; i++)    
    scoreMat[i] = (double *)calloc(tr->contiguousScalingLength, sizeof(double)); 
  
  backtrack = (char **)malloc(unaligned * sizeof(char*));   
  for(i = 0; i < unaligned; i++)    
    backtrack[i] = (char *)calloc(tr->contiguousScalingLength, sizeof(char));   

  gapVector = (double *)calloc(nGaps, sizeof(double));

  gapVector[0] = scoreGeneric(tr->partitionData[0].states, &ref[vectorOffset * 0], gap, diagptable, refExp[0], tr->partitionData[0].dataType);
  for(j = 1; j < nGaps; j++)
    gapVector[j] = gapVector[j-1] + scoreGeneric(tr->partitionData[0].states, &ref[vectorOffset * j], gap, diagptable, refExp[j], tr->partitionData[0].dataType);
   
  scoreMat[0][0] =  scoreGeneric(tr->partitionData[0].states, &ref[vectorOffset * 0], &al[tipOffset * raw[0]], diagptable, refExp[j], tr->partitionData[0].dataType);  
  
  backtrack[0][0] = DIAG;
 
  for(j = 1; j <= nGaps; j++)
    {          
      path1 = gapVector[j-1]   + scoreGeneric(tr->partitionData[0].states, &ref[vectorOffset * j], &al[tipOffset * raw[0]], diagptable, refExp[j], tr->partitionData[0].dataType);
      path2 = scoreMat[0][j-1] + scoreGeneric(tr->partitionData[0].states, &ref[vectorOffset * j], gap, diagptable, refExp[j], tr->partitionData[0].dataType);
      scoreMat[0][j] =  MAX(path1,path2);
     
      if(path1 > path2)
	backtrack[0][j] = DIAG;
      else
	backtrack[0][j] = LEFT;      
    }          

  for(i = 1; i < unaligned; i++)
    for(j = i; j <= i + nGaps; j++)
      {
	assert(scoreMat[i-1][j-1] != 0);
	if(scoreMat[i][j-1] == 0)
	  {
	    scoreMat[i][j] = scoreMat[i-1][j-1] + scoreGeneric(tr->partitionData[0].states, &ref[vectorOffset * j], &al[tipOffset * raw[i]], diagptable, refExp[j], tr->partitionData[0].dataType);	   
	    backtrack[i][j] = DIAG;
	  }
	else
	  {
	    path1 = scoreMat[i-1][j-1] + scoreGeneric(tr->partitionData[0].states, &ref[vectorOffset * j], &al[tipOffset * raw[i]], diagptable, refExp[j], tr->partitionData[0].dataType);
	    path2 = scoreMat[i][j-1]   + scoreGeneric(tr->partitionData[0].states, &ref[vectorOffset * j], gap,                     diagptable, refExp[j], tr->partitionData[0].dataType);
	    scoreMat[i][j] = MAX(path1, path2);

	   
	    if(path1 > path2)
	      backtrack[i][j] = DIAG;
	    else
	      backtrack[i][j] = LEFT;	  	    
	  }	    	
      }
     
  result = scoreMat[unaligned - 1][tr->contiguousScalingLength - 1];

  
  if(result > currentBest)
    {         
      int counterTip =  tr->contiguousScalingLength - 1;	 
	  
      i = unaligned - 1;
      j = tr->contiguousScalingLength - 1;
	  
      while(i >= 0 && j >= 0)
	{
	  assert(backtrack[i][j] != 0);
	  if(backtrack[i][j] == DIAG)
	    {	     
	      
	      tempAlignment[counterTip--]   = raw[i];
	      i--;
	      j--;
	    }
	  else
	    {	      
		  
	      tempAlignment[counterTip--]   = 15;
	      j--;
	    }
	}
	  
      while(j >= 0)
	{	 
	  tempAlignment[counterTip--]   = 15;	 
	  j--;
	}
	  
	  assert(counterTip == -1);      
    }
  
  
  for(i = 0; i < unaligned; i++)
    {
      free(scoreMat[i]);
      free(backtrack[i]);
    }

  free(backtrack);        
  free(scoreMat);   
  free(gapVector);

  return result;
}


void addTraverseRobIterative(tree *tr, boolean withHeuristics)
{      
  int 
    i,
    inserts;

  contiguousBranch
    *b = (contiguousBranch*)NULL;

  double 
    result,
    z[NUM_BRANCHES],
    defaultArray[NUM_BRANCHES];       

  if(withHeuristics)
    assert(tr->heuristicInsertions > 0);

  for(i = 0; i < tr->branchCounter; i++)
    {
      b=&(tr->LOCAL_bInfo[i]);      		 	    	 
       
      assert(!tr->useFastScaling);

      /*printf("%d %d\n", b->branchNumber, tr->threadID);*/

      for(inserts = 0; inserts < tr->numBranches; inserts++) 
	{	      
	  z[inserts] = sqrt(b->branchLengths[inserts]);      
	  defaultArray[inserts] = defaultz;
	  
	  if(z[inserts] < zmin) 
	    z[inserts] = zmin;
	  if(z[inserts] > zmax)
	    z[inserts] = zmax;
	}        
                     
      newviewClassify(tr, b, z);   
        
      for (inserts = 0; inserts < tr->numberOfTipsForInsertion; inserts++) 
	{ 	  
	  if(tr->heuristicInsertions >= 0 || (tr->heuristicInsertions == -1 && b->executeThem[inserts] == 1))
	    {	
	      if(tr->alignQuerySeqs)
		{
		  double *diagptable = makeDiagPtable(tr, defaultz, DNA_DATA, tr->partitionData[0].gammaRates, tr->partitionData[0].EIGN);
		  unsigned char *tempAlignment = (unsigned char*)malloc(sizeof(unsigned char) * tr->contiguousScalingLength);
		  		  		  
		  result = mlAlignGenericContiguous(tr, tr->temporaryVector, tr->temporaryScaling, tr->rawSeqs[inserts].seq, tr->rawSeqs[inserts].length, 
						    diagptable, tr->epaData->scoreVector[inserts], tempAlignment);

		  if(result > tr->epaData->scoreVector[inserts])		 
		    memcpy(tr->contiguousTips[tr->inserts[inserts]], tempAlignment, sizeof(unsigned char) * tr->contiguousScalingLength);
		    
		  free(tempAlignment);
		  free(diagptable);
		}
	      else
		result = evalCL(tr, tr->temporaryVector, tr->temporaryScaling, tr->contiguousTips[tr->inserts[inserts]], defaultArray);
	      
	      if(result > tr->epaData->scoreVector[inserts])
		{	     
		  tr->epaData->scoreVector[inserts]  = result;
		  tr->epaData->branchVector[inserts] = b->branchNumber;
		}
	      
	      if(tr->heuristicInsertions > 0)
		addInsertion(b->branchNumber, result, tr->heuristicInsertions, tr->bBr, inserts); 
	    }
	}	  
    } 
} 





#endif






static void setupBranchMetaInfo(tree *tr, int *countBranches, nodeptr p, int nTips, branchInfo *bInf)
{
  if(isTip(p->number, tr->mxtips))    
    {      
      p->bInf       = &bInf[*countBranches];
      p->back->bInf = &bInf[*countBranches];               	      

      bInf[*countBranches].oP = p;
      bInf[*countBranches].oQ = p->back;
      
      bInf[*countBranches].originalBranchLength = getBranch(tr, p->z, p->back->z);      

      *countBranches =  *countBranches + 1;
      return;
    }
  else
    {
      nodeptr q;
      assert(p == p->next->next->next);

      p->bInf       = &bInf[*countBranches];
      p->back->bInf = &bInf[*countBranches];

      bInf[*countBranches].oP = p;
      bInf[*countBranches].oQ = p->back;

      bInf[*countBranches].originalBranchLength = getBranch(tr, p->z, p->back->z);           
           
      *countBranches =  *countBranches + 1;            

      q = p->next;

      while(q != p)
	{
	  setupBranchMetaInfo(tr, countBranches, q->back, nTips, bInf);	
	  q = q->next;
	}
     
      return;
    }
}
 


static double treeRec(tree *tr, int qNumber, nodeptr p, double currentd, int currenti, double **dMatrix, int **iMatrix)
{
  double branchLength = getBranch(tr, p->z, p->z);

  if(isTip(p->number, tr->mxtips))
    {          
      dMatrix[qNumber][p->number] /*= dMatrix[p->number][qNumber]*/ = currentd + branchLength;
      iMatrix[qNumber][p->number] /*= iMatrix[p->number][qNumber]*/ = currenti + 1;

      return branchLength;
    }
  else
    {
      double 
	acc = 0.0;
      
      nodeptr r = p->next;

      while(r != p)
	{
	  acc += treeRec(tr, qNumber, r->back, currentd + branchLength, currenti + 1, dMatrix, iMatrix);
	  r = r->next;
	}

      return acc + branchLength;
    }
}


static void treeDistances(tree *tr, int *perm, double *dist, double *nodes, double *maxDist, int *maxNodes)
{
  int 
    numTaxa = tr->mxtips + 1,
    **iMat = (int **)malloc(sizeof(int *) * numTaxa),
    i, 
    j,
    iMax = 0,
    meanCount = 0,
    counter = 0,
    permCount = 0;
  
  double 
    **dMat = (double **)malloc(sizeof(double *) * numTaxa),
    length = -1.0, 
    dMax = 0.0,
    meanDist = 0.0,
    meanNodeCount = 0.0;
 
  for(i = 1; i < numTaxa; i++)
    {
      dMat[i] = (double*)calloc(numTaxa, sizeof(double));
      iMat[i] = (int*)calloc(numTaxa, sizeof(int));
    }

  for(i = 1; i < numTaxa; i++)
    {
      if(perm[i] == 1)       	 
	{
	  length = treeRec(tr, tr->nodep[i]->number, tr->nodep[i]->back, 0.0, 0, dMat, iMat);	        
	  permCount++;
	}
    }

    
  for(i = 1; i < numTaxa; i++)
    for(j = 1; j < numTaxa; j++)	 	   
      assert((ABS(dMat[i][j] - dMat[j][i]) < 0.00001) && iMat[i][j] == iMat[j][i]);	  
      
  for(i = 1; i < numTaxa; i++)
    for(j = i + 1; j < numTaxa; j++)
      {
	if(dMax < dMat[i][j])
	  dMax = dMat[i][j];
	
	if(iMax < iMat[i][j])
	  iMax = iMat[i][j];
      }
  
  for(i = 1; i < numTaxa; i++)
    {
      if(perm[i])
	{
	  for(j = i + 1; j < numTaxa; j++)
	    {
	      if(perm[j])
		{
		  meanDist += dMat[i][j];
		  
		  meanCount += (iMat[i][j] - 1);
		  
		  assert(dMat[i][j] > 0.0);
		  assert(iMat[i][j] > 1);		 
		  counter++;
		}
	    }
	}
    }
  
  assert(counter == (((permCount * permCount) - permCount) / 2));

  meanDist      = meanDist / ((double)counter);
  meanNodeCount = ((double)meanCount) / ((double)counter);

  *maxDist = dMax;
  *maxNodes = iMax - 1;

  *dist = meanDist;
  *nodes = meanNodeCount;

  /*
    printf("Ref Tree Length: %f\n", length);  
    printf("Ref Tree Max path length: %f\n", dMax);
    printf("Ref Tree Max path node count: %d\n", *maxNodes);
    printf("Mean pair-wise dist: %f\n", meanDist);
    printf("Mean pair-wise node count: %f\n\n", meanNodeCount);  
  */

  for(i = 1; i < numTaxa; i++)
    {
      free(dMat[i]);
      free(iMat[i]);
    }

  free(dMat);
  free(iMat);
}

static void insertionRec(tree *tr, int qNumber, nodeptr p, double currentd, int currenti, double **dMatrix, int **iMatrix, boolean thorough,
			 nodeptr *insertNodes, int numberOfTipsForInsertion, double *branches)
{  
  if(isTip(p->number, tr->mxtips))         
      return;    
  else
    {         
      nodeptr r = p->next;
      int j;

      while(r != p)
	{
	  /* branch p->next -- p->next-Back */
	  double branchLength = getBranch(tr, r->z, r->z);

	  insertionRec(tr, qNumber, r->back, currentd + branchLength, currenti + 1, dMatrix, iMatrix, thorough, insertNodes, 
		       numberOfTipsForInsertion, branches);

	  for(j = 0; j < numberOfTipsForInsertion; j++)
	    {
	      if(qNumber != j && (r == insertNodes[j] || r->back == insertNodes[j]))
		{
		  iMatrix[qNumber][j] = currenti;

		  if(thorough)		
		    dMatrix[qNumber][j] = currentd + (branchLength * 0.5) + branches[j];
		  else
		    dMatrix[qNumber][j] = currentd + (branchLength * 0.5);
		}
	    }

	  r = r->next;
	}

      return;
    }
}


static void insertionDistances(tree *tr, int numberOfTipsForInsertion, boolean thorough, double *branches, nodeptr *insertNodes, double *dist, double *nodes,
			       double *maxDist, int *maxNodes)
{
  int    
    **iMat = (int **)malloc(sizeof(int *) * numberOfTipsForInsertion),
    i, 
    j,
    iMax = 0,
    meanCount = 0,
    counter = 0;
  
  double 
    **dMat = (double **)malloc(sizeof(double *) * numberOfTipsForInsertion),
    dMax = 0.0,
    meanDist = 0.0,
    meanNodeCount = 0.0;
 
  for(i = 0; i < numberOfTipsForInsertion; i++)
    {
      dMat[i] = (double*)calloc(numberOfTipsForInsertion, sizeof(double));
      iMat[i] = (int*)calloc(numberOfTipsForInsertion, sizeof(int));
    }

  for(i = 0; i < numberOfTipsForInsertion; i++)
    {     
      nodeptr 
	p = insertNodes[i];           

      double x = getBranch(tr, p->z, p->z) * 0.5;
      
      if(thorough)
	x += branches[i];
     
      for(j = 0; j < numberOfTipsForInsertion; j++)
	{
	  if(i != j && (p == insertNodes[j] || p->back == insertNodes[j]))
	    {
	      iMat[i][j] = 0;

	      if(thorough)		
		dMat[i][j] = ABS(branches[i] - branches[j]);
	      else
		dMat[i][j] = 0.0;
	    }
	}
      
      if(!isTip(p->number, tr->mxtips))
	insertionRec(tr, i, p, x, 1, dMat, iMat, thorough, insertNodes, numberOfTipsForInsertion, branches);	
      if(!isTip(p->back->number, tr->mxtips))
	insertionRec(tr, i, p->back, x, 1, dMat, iMat, thorough, insertNodes, numberOfTipsForInsertion, branches);	      	
    }

 
  
  for(i = 0; i < numberOfTipsForInsertion; i++)
    for(j = 0; j < numberOfTipsForInsertion; j++)	 	  
      assert((ABS(dMat[i][j] - dMat[j][i]) < 0.00001) && iMat[i][j] == iMat[j][i]);	 
  
  
  for(i = 0; i < numberOfTipsForInsertion; i++)
    for(j = i + 1; j < numberOfTipsForInsertion; j++)
      {
	if(dMax < dMat[i][j])
	  dMax = dMat[i][j];
	
	if(iMax < iMat[i][j])
	  iMax = iMat[i][j];	
      }

 

  for(i = 0; i < numberOfTipsForInsertion; i++)
    for(j = i + 1; j < numberOfTipsForInsertion; j++)
      {
	meanDist += dMat[i][j];
		  
	meanCount += (iMat[i][j] - 1);
		  	
	assert(dMat[i][j] >= 0.0);
	assert(iMat[i][j] >= 0);		 
	counter++;
      }

  assert(counter == (((numberOfTipsForInsertion * numberOfTipsForInsertion) - numberOfTipsForInsertion) / 2));
   
  meanDist      = meanDist / ((double)counter);
  meanNodeCount = ((double)meanCount) / ((double)counter);

  /*
    printf("Query Max path length: %f\n", dMax);
    printf("Query Max path node count: %d\n", iMax);
    printf("Query Mean pair-wise dist: %f\n", meanDist);
    printf("Query Mean pair-wise node count: %f\n", meanNodeCount);
  */
  


  *dist = meanDist;
  *nodes = meanNodeCount;

  *maxDist  = dMax;
  *maxNodes = iMax;

  
  for(i = 0; i < numberOfTipsForInsertion; i++)
    {
      free(dMat[i]);
      free(iMat[i]);
    }

  free(dMat);
  free(iMat);
}



static void diversityStats(tree *tr, int numberOfTipsForInsertion, boolean thorough, double *branches, nodeptr *insertNodes, int *perm, 
			   double *queryDist, double *queryNodes, double *maxQueryDist, int *maxQueryNodes, 
			   double *treeDist,  double *treeNodes,  double *maxTreeDist,  int *maxTreeNodes, int replicate)
{
  double 
    thisTreeDist = 0.0,
    thisQueryDist = 0.0,
    thisMaxTreeDist = 0.0,
    thisMaxQueryDist = 0.0,
    thisTreeNodes = 0.0,
    thisQueryNodes = 0.0;

  int     
    thisMaxTreeNodes = 0,
    thisMaxQueryNodes = 0;

  treeDistances(tr, perm, &thisTreeDist, &thisTreeNodes, &thisMaxTreeDist, &thisMaxTreeNodes);

  insertionDistances(tr, numberOfTipsForInsertion, thorough, branches, insertNodes, &thisQueryDist, &thisQueryNodes, &thisMaxQueryDist, &thisMaxQueryNodes);

  if(replicate >= 0)
    {
      printBothOpen("Diversity Stats for this replicate\n", replicate);
      printBothOpen("Max Query Distance:  %f\n", thisMaxQueryDist);
      printBothOpen("Max Tree Distance:   %f\n", thisTreeDist);
      printBothOpen("Mean Query Distance: %f\n", thisQueryDist);
      printBothOpen("Mean Tree Distance:  %f\n", thisTreeDist);

      printBothOpen("Max Query  Count: %d\n", thisMaxQueryNodes);
      printBothOpen("Max Tree   Count: %d\n", thisMaxTreeNodes);
      printBothOpen("Mean Query Count: %f\n", thisQueryNodes);
      printBothOpen("Mean Tree  Count: %f\n", thisTreeNodes);
    }

  *queryDist += thisQueryDist;
  *queryNodes += thisQueryNodes;
  *maxQueryDist += thisMaxQueryDist;
  *maxQueryNodes += thisMaxQueryNodes;

  *treeDist += thisTreeDist;    
  *treeNodes += thisTreeNodes;
  *maxTreeDist += thisMaxTreeDist;
  *maxTreeNodes += thisMaxTreeNodes;  
}



static void allocbBr(tree *tr, int numberOfTipsForInsertion)
{
  if(tr->heuristicInsertions > 0)
    {
#ifdef _USE_PTHREADS
      masterBarrier(THREAD_ALLOC_BR, tr);
#else
      int 
	i,
	j;
      
      tr->bBr = (bestBranches *)malloc(sizeof(bestBranches) * numberOfTipsForInsertion);
      for(i = 0; i < numberOfTipsForInsertion; i++)
	{
	  tr->bBr[i].parsimonyScores = (unsigned int*)    malloc(sizeof(unsigned int)      * tr->heuristicInsertions);
	  tr->bBr[i].likelihoods = (double *)    malloc(sizeof(double)      * tr->heuristicInsertions);
	  tr->bBr[i].branch      = (int *)malloc(sizeof(int*) * tr->heuristicInsertions);
	  tr->bBr[i].count       = 0;
	  tr->bBr[i].minLH       = -1;
	  for(j = 0; j < tr->heuristicInsertions; j++)
	    {
	      tr->bBr[i].parsimonyScores[j] = INT_MAX;
	      tr->bBr[i].likelihoods[j] = unlikely;
	      tr->bBr[i].branch[j]      = -1;
	    }
	}
#endif
    }
}

#ifdef _USE_PTHREADS

typedef struct
{
  int parsimonyScore;
  int branchNumber;
  double likelihoodScore;
} branchHeuristics;





static int branchHeuristicsCompareLH(const void *p1, const void *p2)
{
  branchHeuristics *rc1 = (branchHeuristics *)p1;
  branchHeuristics *rc2 = (branchHeuristics *)p2;

  double i = rc1->likelihoodScore;
  double j = rc2->likelihoodScore;

  if (i > j)
    return (-1);
  if (i < j)
    return (1);
  return (0);
}

static int branchHeuristicsCompareMP(const void *p1, const void *p2)
{
  branchHeuristics *rc1 = (branchHeuristics *)p1;
  branchHeuristics *rc2 = (branchHeuristics *)p2;

  int i = rc1->parsimonyScore;
  int j = rc2->parsimonyScore;

  if (i > j)
    return (1);
  if (i < j)
    return (-1);
  return (0);
}
#endif


static void consolidateAndFreeBr(tree *tr, int numberOfTipsForInsertion)
{
   if(tr->heuristicInsertions > 0)
     {
#ifdef _USE_PTHREADS
       {
	 int
	   globalCount,
	   tid,
	   i, 
	   j;
	 
	 branchHeuristics *bh = (branchHeuristics *)malloc(sizeof(branchHeuristics) * NumberOfThreads *  tr->heuristicInsertions);
	   	

	 for(i = 0; i < numberOfTipsForInsertion; i++)	
	   {	 	        
	     for(tid = 0, globalCount = 0; tid < NumberOfThreads; tid++)
	       {	     
		 bestBranches *bBr = (bestBranches *)branchData[tid];
		 
		 for(j = 0; j < bBr[i].count; j++, globalCount++)
		   {
		     bh[globalCount].parsimonyScore  = bBr[i].parsimonyScores[j];
		     bh[globalCount].likelihoodScore = bBr[i].likelihoods[j];
		     bh[globalCount].branchNumber = bBr[i].branch[j];		       
		   }		   
	       }
	
	     /*printf("%d %d\n", globalCount, tr->heuristicInsertions);*/
	     assert(globalCount >= tr->heuristicInsertions);	       
	     
	     

	     if(tr->fastEPA_MP)	       		
	       qsort(bh, globalCount, sizeof(branchHeuristics), branchHeuristicsCompareMP);		
	     else
	       qsort(bh, globalCount, sizeof(branchHeuristics), branchHeuristicsCompareLH);	    
	    
	     /*printf("Insert %d\n", i);*/
	     for(j = 0; j < tr->heuristicInsertions; j++)	       		 	      
	       {
		 tr->bInf[bh[j].branchNumber].executeThem[i] = 1;			    	     	     	      
		 /*printf("%f ", bh[j].likelihoodScore); */
	       }
	     /* printf("\n"); */
	   }

	 free(bh);

	 masterBarrier(THREAD_FREE_BRANCH_INFO, tr);
       }
#else
       {
	 int 
	   i, 
	   j;
	 
	 for(i = 0; i < numberOfTipsForInsertion; i++)	
	   {
	     assert(tr->bBr[i].count == tr->heuristicInsertions);

	     printf("Insert %d\n", i);

	     for(j = 0; j < tr->bBr[i].count; j++)	       
	       {
		 tr->bInf[tr->bBr[i].branch[j]].executeThem[i] = 1;		  		 		
		 printf("%f ", tr->bBr[i].likelihoods[j]);
	       }
	     printf("\n");
		
	   }
	 
	 for(i = 0; i < numberOfTipsForInsertion; i++)
	   {
	     free(tr->bBr[i].parsimonyScores);
	     free(tr->bBr[i].likelihoods);
	     free(tr->bBr[i].branch);
	   }
	 free(tr->bBr);		 
       }
#endif
     }	
}


static void initBranchInfo(tree *tr, int  numberOfTipsForInsertion, double *branches, double *insertLikelihoods, nodeptr *insertNodes)
{
  int j;
  
  for(j = 0; j < numberOfTipsForInsertion; j++) 
    {
      branches[j]          = -1.0;
      insertLikelihoods[j] = unlikely;
      insertNodes[j]     = (nodeptr)NULL;
    }  

#ifdef _USE_PTHREADS
  masterBarrier(THREAD_INIT_BRANCH_INFO, tr);  
#endif
}

#ifdef _USE_PTHREADS

static void collectAlignmentMultiGrain(tree *tr, int numberOfTipsForInsertion, double *insertLikelihoods)
{    
  int tid;                 
  
  for(tid = 0; tid < NumberOfThreads; tid++)
    {
      int k;
      threadEPAdata *epd = (threadEPAdata *)epaData[tid];
      
      for(k = 0; k < numberOfTipsForInsertion; k++)
	{	     						
	  if(epd->scoreVector[k] >  insertLikelihoods[k])		 		   		    		  
	    insertLikelihoods[k] = epd->scoreVector[k];		  
	}
    }	
  
  for(tid = 0; tid < NumberOfThreads; tid++)
    {
      int k;
      threadEPAdata *epd = (threadEPAdata *)epaData[tid];
      
      for(k = 0; k < numberOfTipsForInsertion; k++)
	{	       	
	  if(epd->scoreVector[k] == insertLikelihoods[k])
	    {	
	      unsigned char ***source = (unsigned char ***)globalContiguousTips;
	      
	      memcpy(tr->yVector[tr->inserts[k]], source[tid][tr->inserts[k]], sizeof(unsigned char) * tr->contiguousScalingLength);
	    }
	}
    }
}     

#endif 



static void linkResults2Tree(tree *tr, int numberOfTipsForInsertion, nodeptr *insertNodes, double *branches, boolean thorough, double *insertLikelihoods, int *insertNumbers)
{
#ifdef _USE_PTHREADS
  {  
    int tid;                 

    for(tid = 0; tid < NumberOfThreads; tid++)
      {
	int k;
	threadEPAdata *epd = (threadEPAdata *)epaData[tid];
	
	for(k = 0; k < numberOfTipsForInsertion; k++)
	  {	     						
	    if(epd->scoreVector[k] >  insertLikelihoods[k])		 		   		    		  
	      insertLikelihoods[k] = epd->scoreVector[k];		  
	  }
      }	
    
    for(tid = 0; tid < NumberOfThreads; tid++)
      {
	int k;
	threadEPAdata *epd = (threadEPAdata *)epaData[tid];
	
	for(k = 0; k < numberOfTipsForInsertion; k++)
	  {	       	
	    if(epd->scoreVector[k] == insertLikelihoods[k])
	      {
		double branch = epd->insertionBranchLengths[k];
		branchInfo *b = &(tr->bInf[epd->branchVector[k]]);				

		insertNumbers[k] = epd->branchVector[k];			       

		insertNodes[k] = b->oP;

		/*printf("%d %f %s\n", k, insertLikelihoods[k], b->branchLabel);*/
						
		if(thorough)
		  {		   
		    branches[k] = branch;
		    b->branches[k] = b->branches[k] + branch;
		  }
		
		b->countThem[k] = b->countThem[k] + 1;		  
	      }
	  }
      }	   
  }     
#else
  {
    int j;

    for(j = 0; j < numberOfTipsForInsertion; j++)
      {
	assert(insertNodes[j]);
	insertNodes[j]->bInf->countThem[j] = insertNodes[j]->bInf->countThem[j] + 1;

	/*printf("%d %f %s\n", j, insertLikelihoods[j], insertNodes[j]->bInf->branchLabel);*/

	if(thorough)
	  {
	    insertNodes[j]->bInf->branches[j]  = insertNodes[j]->bInf->branches[j] + branches[j];
	    assert(branches[j] != -1);
	  }
      }     
  }
#endif
}


static void printReplicate(tree *tr, int numberOfTipsForInsertion, double *branches, boolean thorough, int replicates, int *inserts, nodeptr *insertNodes, int *insertNumbers)
{
  char 
    fileName[2048], 
    buf[64];
  
  FILE *f;

  int j;
  
  strcpy(fileName, "RAxML_classificationReplicate.");
  sprintf(buf, "%d", replicates);
  strcat(fileName, buf);
  strcat(fileName,".");
  strcat(fileName, run_id);	    	    
	    
  f = fopen(fileName, "w");
  
  for(j = 0; j < numberOfTipsForInsertion; j++)
    {		     
#ifdef _USE_PTHREADS
      if(thorough)
	fprintf(f, "%s %s 1 %f\n", tr->nameList[inserts[j]], tr->bInf[insertNumbers[j]].branchLabel, branches[j]);
      else
	fprintf(f, "%s %s 1\n", tr->nameList[inserts[j]], tr->bInf[insertNumbers[j]].branchLabel);
#else
      if(thorough)
	fprintf(f, "%s %s 1 %f\n", tr->nameList[inserts[j]], insertNodes[j]->bInf->branchLabel, branches[j]);
      else
	fprintf(f, "%s %s 1\n", tr->nameList[inserts[j]], insertNodes[j]->bInf->branchLabel);
#endif
    }
  
  fclose(f);  
}

#ifdef _USE_PTHREADS

static void setAlign(tree *tr, boolean value)
{
  tr->alignQuerySeqs = value;
  masterBarrier(THREAD_SET_ALIGN, tr);

  return;
}


static void setupMultiGrain(tree *tr, int numberOfBranches, int *inserts, int numberOfTipsForInsertion)
{
  tr->contiguousVectorLength = getContiguousVectorLength(tr);
  tr->contiguousScalingLength = getContiguousScalingLength(tr);          
  
  tr->numberOfBranches = numberOfBranches;
  
  tr->inserts                  = inserts;
  tr->numberOfTipsForInsertion = numberOfTipsForInsertion;        
  
  tr->GLOBAL_bInfo = (contiguousBranch*)malloc(sizeof(contiguousBranch));
  
  tr->GLOBAL_bInfo->leftNodeNumber  = -1;
  tr->GLOBAL_bInfo->rightNodeNumber = -1;
  
  tr->GLOBAL_bInfo->left  = (double*)malloc_aligned(sizeof(double) * tr->contiguousVectorLength);
  tr->GLOBAL_bInfo->leftScaling = (int*)malloc(sizeof(int) * tr->contiguousScalingLength);    
  
  tr->GLOBAL_bInfo->right = (double*)malloc_aligned(sizeof(double)  * tr->contiguousVectorLength);
  tr->GLOBAL_bInfo->rightScaling = (int*)malloc(sizeof(int) * tr->contiguousScalingLength);
  
  tr->GLOBAL_bInfo->leftParsimony  = (parsimonyVector *)tr->GLOBAL_bInfo->left;
  tr->GLOBAL_bInfo->rightParsimony = (parsimonyVector *)tr->GLOBAL_bInfo->right;

  masterBarrier(THREAD_COPY_BRANCHINFO_POINTER, tr);             

  return;
}
#endif

static void setupBranchInfo(tree *tr, nodeptr q, int numberOfBranches)
{
  int countBranches = 0;

  setupBranchMetaInfo(tr, &countBranches, q, tr->ntips, tr->bInf);
    
  assert(countBranches == numberOfBranches);
}


static void aligner(tree *tr, analdef *adef)
{
  unsigned char 
    undetermined;
  int    
    i, 
    j,  
    *perm, 
    *inserts, 
    numberOfTipsForInsertion = 0,    
    numberOfBranches = 2 * tr->ntips - 3,
    dataType,       
    *insertNumbers;  
  
  double  
    *diagptable = (double *)NULL,
    *insertLikelihoods,
    *branches;
  
  nodeptr 
    *insertNodes, 
    r, 
    q;          
  
  char    
    alignmentFileName[1024];

  FILE    
    *alignmentFile;


  adef->outgroup = FALSE;
  tr->doCutoff   = FALSE;

#ifdef _USE_PTHREADS
  setAlign(tr, TRUE);
#endif

  assert(adef->restart);

  dataType = tr->partitionData[0].dataType;

  assert(tr->NumberOfModels == 1 && (tr->rateHetModel == GAMMA || tr->rateHetModel == GAMMA_I));

  printBothOpen("\nRAxML Alignment Algorithm using %s criterion\n", (adef->dynamicAlignment == ALIGN_ML)?"ML":"MP");

  if(adef->dynamicAlignment == ALIGN_ML)
    {
      evaluateGenericInitrav(tr, tr->start);
      modOpt(tr, adef, TRUE, 1.0); 
#ifdef _USE_PTHREADS
      masterBarrier(THREAD_COPY_FRAC_CHANGE, tr);
#endif
      
      switch(tr->rateHetModel)
	{
	case GAMMA:
	  printBothOpen("GAMMA based Likelihood score of reference tree: %f\n\n", tr->likelihood);
	  break;
	case GAMMA_I:
	  printBothOpen("GAMMA+P-Invar based Likelihood score of reference tree: %f\n\n", tr->likelihood);
	  break; 
	case CAT:
	  printBothOpen("CAT based Likelihood score of reference tree: %f\n\n", tr->likelihood);
	  break;
	default:
	  assert(0);
	}

    }

  perm    = (int *)calloc(tr->mxtips + 1, sizeof(int));
  inserts = (int *)calloc(tr->mxtips, sizeof(int));

  markTips(tr->start,       perm, tr->mxtips);
  markTips(tr->start->back, perm ,tr->mxtips);  

  for(i = 1; i <= tr->mxtips; i++)
    {
      if(perm[i] == 0)
	{
	  inserts[numberOfTipsForInsertion] = i;
	  numberOfTipsForInsertion++;
	}
    }        

  tr->rawSeqs = (unalignedSequences*)malloc(sizeof(unalignedSequences) * numberOfTipsForInsertion);


  undetermined = getUndetermined(dataType);

  

  for(i = 0; i < numberOfTipsForInsertion; i++)
    {    
      int unaligned = 0;
      unsigned char 
	*tip = tr->yVector[inserts[i]], 
	*raw;
      
      tr->rawSeqs[i].seq = raw = (unsigned char *)malloc(sizeof(unsigned char) * tr->cdta->endsite);                      
	 	  	  
      for(j = 0; j < tr->cdta->endsite; j++)		   	   	    
	if(tip[j] != undetermined)	    	      
	  raw[unaligned++] = tip[j];	      
	 
      tr->rawSeqs[i].length = unaligned;

      if(unaligned == tr->cdta->endsite)
	{
	  printBothOpen("Query sequence %s you want to dynamically align to the reference alignment\n", tr->nameList[inserts[i]]);
	  printBothOpen("does not contain any gaps, i.e., there is nothing to align here, exiting\n");
	  exit(0);
	}
    }

  branches          = (double*)malloc(sizeof(double) *  numberOfTipsForInsertion);
  insertLikelihoods = (double*)malloc(sizeof(double) *  numberOfTipsForInsertion);
  insertNodes       = (nodeptr*)malloc(sizeof(nodeptr) *  numberOfTipsForInsertion);      
  insertNumbers     = (int*)malloc(sizeof(int) *  numberOfTipsForInsertion);      

  tr->bInf              = (branchInfo*)malloc(numberOfBranches * sizeof(branchInfo));

  for(i = 0; i < numberOfBranches; i++)
    {      
      tr->bInf[i].countThem   = (int*)calloc(numberOfTipsForInsertion, sizeof(int));      
      tr->bInf[i].executeThem = (int*)calloc(numberOfTipsForInsertion, sizeof(int));
      tr->bInf[i].branches    = (double*)calloc(numberOfTipsForInsertion, sizeof(double));
      tr->bInf[i].branchNumber = i;
      sprintf(tr->bInf[i].branchLabel, "I%d", i);
    }
     
  printBothOpen("RAxML will align %d Query Sequences into the %d branches of the reference tree with %d taxa\n\n",  numberOfTipsForInsertion, (2 * tr->ntips - 3), tr->ntips);

  r = tr->nodep[(tr->nextnode)++];     

  q = findAnyTip(tr->start, tr->rdta->numsp);
  assert(isTip(q->number, tr->rdta->numsp));
  assert(!isTip(q->back->number, tr->rdta->numsp));
	 
  q = q->back;               

#ifdef _USE_PTHREADS
  setupMultiGrain(tr, numberOfBranches, inserts, numberOfTipsForInsertion);
#endif
  setupBranchInfo(tr, q, numberOfBranches);   
    
  if(adef->dynamicAlignment == ALIGN_ML) 
    {
      diagptable = makeDiagPtable(tr, defaultz, dataType, tr->partitionData[0].gammaRates,tr->partitionData[0].EIGN);  
      evaluateGenericInitrav(tr, q->back);
    }
  
  initBranchInfo(tr, numberOfTipsForInsertion, branches, insertLikelihoods, insertNodes);

  assert(!(tr->fastEPA_MP || tr->fastEPA_ML));
  
  tr->heuristicInsertions = 0;
#ifdef _USE_PTHREADS
  masterBarrier(THREAD_EPA_SET_HEURISTICS, tr);
#endif

  switch(adef->dynamicAlignment)
    {
    case ALIGN_MP:	      
      makeParsimonyInsertions(tr, q, r, inserts, numberOfTipsForInsertion, tr->heuristicInsertions, tr->bBr, tr->rawSeqs, TRUE);		  
      break;
    case ALIGN_ML:
#ifdef _USE_PTHREADS
      if(adef->thoroughInsertion)
	assert(0); /* this will not be implemented right now at last */
      else
	{	 
	  collectLikelihoodVectors(tr, q, numberOfBranches);
	  masterBarrier(THREAD_INSERT_CLASSIFY, tr);
	}
#else	 		 	  
      addTraverseRob(tr, r, q, inserts,  numberOfTipsForInsertion, insertLikelihoods, branches,
		     insertNodes, TRUE, diagptable, TRUE, tr->rawSeqs, adef->thoroughInsertion, dataType, tr->bBr, TRUE);       
#endif
      break;
    default:
      assert(0);
    }   

#ifdef _USE_PTHREADS
  collectAlignmentMultiGrain(tr, numberOfTipsForInsertion, insertLikelihoods);
#endif
  
  strcpy(alignmentFileName,            workdir);
  strcat(alignmentFileName,            "RAxML_alignment.");
  strcat(alignmentFileName,            run_id);

  alignmentFile = myfopen(alignmentFileName, "w");
      
  assert(tr->cdta->endsite == tr->originalCrunchedLength);                             

  fprintf(alignmentFile, "%d %d\n", tr->mxtips, tr->originalCrunchedLength);

  for(i = 1; i <= tr->mxtips; i++)
    {
      unsigned char *tipI = tr->yVector[i];

      fprintf(alignmentFile, "%s ", tr->nameList[i]); 

      for(j = 0; j < tr->originalCrunchedLength; j++)	
	fprintf(alignmentFile, "%c", getInverseMeaning(dataType, tipI[j]));
	         
      fprintf(alignmentFile, "\n");
    }

  fclose(alignmentFile);
         
  printBothOpen("Alignment written to file: %s\n\n", alignmentFileName);

  exit(0);
}

static void classifier(tree *tr, analdef *adef)
{
  int
    root = -1,
    i, 
    j,  
    *perm, 
    *inserts, 
    numberOfTipsForInsertion = 0,    
    numberOfBranches = 2 * tr->ntips - 3,
    maxTreeNodes = 0,
    maxQueryNodes = 0,
    *insertNumbers;  
  
  double     
    *insertLikelihoods,
    *branches,
    treeDist = 0.0,
    queryDist = 0.0,
    maxTreeDist = 0.0,
    maxQueryDist = 0.0,
    treeNodes = 0,
    queryNodes = 0;
  
  nodeptr 
    *insertNodes, 
    r, 
    q;    
     
  boolean     
    thorough       = adef->thoroughInsertion;

  char 
    labelledTreeFileName[1024],
    originalLabelledTreeFileName[1024],
    classificationFileName[1024];

  FILE 
    *treeFile, 
    *classificationFile;

#ifdef _USE_PTHREADS
  setAlign(tr, FALSE);
#endif

  adef->outgroup = FALSE;
  tr->doCutoff   = FALSE;

  assert(adef->restart);

  tr->bBr = (bestBranches *)NULL;
  
  printBothOpen("\nRAxML Evolutionary Placement Algorithm\n");

  printBothOpen("Thorough  Insertion Method with branch length optimization: %s\n", thorough?"YES":"NO");
  printBothOpen("Bootstrap insertions with heuristics\n\n"); 

  evaluateGenericInitrav(tr, tr->start); 
  modOpt(tr, adef, TRUE, 1.0);

#ifdef _USE_PTHREADS
  if(thorough)
    masterBarrier(THREAD_COPY_FRAC_CHANGE, tr);
#endif

  perm    = (int *)calloc(tr->mxtips + 1, sizeof(int));
  inserts = (int *)calloc(tr->mxtips, sizeof(int));

  markTips(tr->start,       perm, tr->mxtips);
  markTips(tr->start->back, perm ,tr->mxtips);

  numberOfTipsForInsertion = 0;

  for(i = 1; i <= tr->mxtips; i++)
    {
      if(perm[i] == 0)
	{
	  inserts[numberOfTipsForInsertion] = i;
	  numberOfTipsForInsertion++;
	}
    }      
  
  printBothOpen("RAxML will place %d Query Sequences into the %d branches of the reference tree with %d taxa\n\n",  numberOfTipsForInsertion, (2 * tr->ntips - 3), tr->ntips);

  if(adef->rapidBoot)
    printBothOpen("RAxML will execute %d Bootstrap replicates to infer classification support\n\n",  adef->multipleRuns);

  assert(numberOfTipsForInsertion == (tr->mxtips - tr->ntips));

  branches          = (double*)malloc(sizeof(double) *  numberOfTipsForInsertion);
  insertLikelihoods = (double*)malloc(sizeof(double) *  numberOfTipsForInsertion);
  insertNodes       = (nodeptr*)malloc(sizeof(nodeptr) *  numberOfTipsForInsertion);      
  insertNumbers     = (int*)malloc(sizeof(int) *  numberOfTipsForInsertion);      

  tr->bInf              = (branchInfo*)malloc(numberOfBranches * sizeof(branchInfo));

  for(i = 0; i < numberOfBranches; i++)
    {      
      tr->bInf[i].countThem   = (int*)calloc(numberOfTipsForInsertion, sizeof(int));      
      tr->bInf[i].executeThem = (int*)calloc(numberOfTipsForInsertion, sizeof(int));
      tr->bInf[i].branches    = (double*)calloc(numberOfTipsForInsertion, sizeof(double));
      tr->bInf[i].branchNumber = i;
      sprintf(tr->bInf[i].branchLabel, "I%d", i);
    }
     

  r = tr->nodep[(tr->nextnode)++];     

  q = findAnyTip(tr->start, tr->rdta->numsp);
  assert(isTip(q->number, tr->rdta->numsp));
  assert(!isTip(q->back->number, tr->rdta->numsp));
	 
  q = q->back;                
#ifdef _USE_PTHREADS
  setupMultiGrain(tr, numberOfBranches, inserts, numberOfTipsForInsertion);
#endif 
  setupBranchInfo(tr, q, numberOfBranches);   
  
  initBranchInfo(tr, numberOfTipsForInsertion, branches, insertLikelihoods, insertNodes);
  
  if(tr->fastEPA_MP || tr->fastEPA_ML)
    {	 	
      tr->heuristicInsertions = MAX(5, (int)(0.5 + (double)numberOfBranches * tr->fastEPAthreshold));
	 
#ifdef _USE_PTHREADS
      masterBarrier(THREAD_EPA_SET_HEURISTICS, tr);
#endif
      
      thorough = FALSE;   
	    
      allocbBr(tr, numberOfTipsForInsertion);	 
	  
      if(tr->fastEPA_MP)
	{ 
	  printBothOpen("Searching for %d out of %d most promising branches with MP heuristics\n", tr->heuristicInsertions, numberOfBranches);	 
   
	  makeParsimonyInsertions(tr, q, r, inserts, numberOfTipsForInsertion, tr->heuristicInsertions, tr->bBr, (unalignedSequences *)NULL, FALSE);
	  
	  evaluateGenericInitrav(tr, q->back);

#ifdef _USE_PTHREADS
	  masterBarrier(THREAD_RESET_GATHER_VALUES, tr);	  
	  collectLikelihoodVectors(tr, q, numberOfBranches);	
#endif	  
	}
	  
      if(tr->fastEPA_ML)
	{
	  printBothOpen("Searching for %d out of %d most promising branches with ML heuristics\n", tr->heuristicInsertions, numberOfBranches);	      
#ifdef _USE_PTHREADS
	  collectLikelihoodVectors(tr, q, numberOfBranches);
	  masterBarrier(THREAD_INSERT_CLASSIFY_FAST_EPA_ML, tr);               			 
#else  		
	  addTraverseRob(tr, r, q, inserts,  numberOfTipsForInsertion, insertLikelihoods, branches,
			 insertNodes, FALSE, (double*)NULL, FALSE, tr->rawSeqs, thorough, -1, tr->bBr, TRUE);
#endif
	}
      
      consolidateAndFreeBr(tr, numberOfTipsForInsertion);

      tr->heuristicInsertions = -1; 

#ifdef _USE_PTHREADS
      masterBarrier(THREAD_EPA_SET_HEURISTICS, tr);
#endif
      
      thorough = TRUE;

#ifdef _USE_PTHREADS
      masterBarrier(THREAD_INSERT_CLASSIFY_THOROUGH, tr);	      
#else     
      addTraverseRob(tr, r, q, inserts,  numberOfTipsForInsertion, insertLikelihoods, branches,
		     insertNodes, FALSE, (double*)NULL, FALSE, tr->rawSeqs, thorough, -1, tr->bBr, TRUE);
#endif
      
      linkResults2Tree(tr, numberOfTipsForInsertion, insertNodes, branches, thorough, insertLikelihoods, insertNumbers);
     
      diversityStats(tr, numberOfTipsForInsertion, thorough, branches, insertNodes, perm, 
		     &queryDist, &queryNodes, &maxQueryDist, &maxQueryNodes, 
		     &treeDist,  &treeNodes,  &maxTreeDist,  &maxTreeNodes, -1);
    }
  else
    {      
      if(adef->rapidBoot)	
	tr->heuristicInsertions = MAX(5, (int)(numberOfBranches/10));/*MAX(5, (int)(0.5 + (double)numberOfBranches * tr->fastEPAthreshold));*/
      else
	tr->heuristicInsertions = 0;
	  
#ifdef _USE_PTHREADS
      masterBarrier(THREAD_EPA_SET_HEURISTICS, tr);
#endif
     
      if(tr->heuristicInsertions > 0)
	allocbBr(tr, numberOfTipsForInsertion);           	      
      
#ifdef _USE_PTHREADS
      collectLikelihoodVectors(tr, q, numberOfBranches);
      if(thorough)	   	    
	masterBarrier(THREAD_INSERT_CLASSIFY_THOROUGH, tr);	    
      else
	masterBarrier(THREAD_INSERT_CLASSIFY, tr);
#else            
      if(adef->rapidBoot)
	addTraverseRob(tr, r, q, inserts,  numberOfTipsForInsertion, insertLikelihoods, (double*)NULL,
		       insertNodes, FALSE, (double*)NULL, FALSE, (unalignedSequences*)NULL, 
		       thorough, -1, tr->bBr, FALSE);	    
      else
	addTraverseRob(tr, r, q, inserts,  numberOfTipsForInsertion, insertLikelihoods, branches,
		       insertNodes, FALSE, (double*)NULL, FALSE, (unalignedSequences*)NULL, 
		       thorough, -1, tr->bBr, TRUE);	
#endif
      
      if(adef->rapidBoot)
	{
	  int 
	    replicates,
	    *originalRateCategories = (int*)malloc(tr->cdta->endsite * sizeof(int)),
	    *originalInvariant      = (int*)malloc(tr->cdta->endsite * sizeof(int));
       
	  assert(tr->heuristicInsertions > 0);

	  memcpy(originalRateCategories, tr->cdta->rateCategory, sizeof(int) * tr->cdta->endsite);
	  memcpy(originalInvariant,      tr->invariant,          sizeof(int) * tr->cdta->endsite); 	 	  
	  
	  consolidateAndFreeBr(tr, numberOfTipsForInsertion);	  

	  if(tr->heuristicInsertions > 0)
	    tr->heuristicInsertions = -1;
	  else
	    tr->heuristicInsertions = 0;     

#ifdef _USE_PTHREADS
	  masterBarrier(THREAD_EPA_SET_HEURISTICS, tr);
#endif
	  
	  for(replicates = 0; replicates < adef->multipleRuns; replicates++)
	    {
	      double rt = gettime();

	      computeNextReplicate(tr, &adef->rapidBoot, originalRateCategories, originalInvariant, TRUE);
	      
#ifdef _USE_PTHREADS
	      masterBarrier(THREAD_CONTIGUOUS_REPLICATE, tr); 
#endif	  	 
	      resetBranches(tr);
	      evaluateGenericInitrav(tr, q->back);	  
	      treeEvaluate(tr, 1);
	      
#ifdef _USE_PTHREADS
	      masterBarrier(THREAD_RESET_GATHER_VALUES, tr);
	      collectLikelihoodVectors(tr, q, numberOfBranches);	     
#endif
	      
	      initBranchInfo(tr, numberOfTipsForInsertion, branches, insertLikelihoods, insertNodes);
	  
#ifdef _USE_PTHREADS
	      if(thorough)	   	    
		masterBarrier(THREAD_INSERT_CLASSIFY_THOROUGH, tr);	    
	      else
		masterBarrier(THREAD_INSERT_CLASSIFY, tr);	  
#else
	      addTraverseRob(tr, r, q, inserts,  numberOfTipsForInsertion, insertLikelihoods, branches,
			     insertNodes, FALSE /*align we set align to FALSE here*/, 
			     (double*)NULL, FALSE, (unalignedSequences*)NULL /*rawSeqs*/, 
			     thorough, -1, tr->bBr, TRUE); 
#endif  
	 
	      linkResults2Tree(tr, numberOfTipsForInsertion, insertNodes, branches, thorough, insertLikelihoods, insertNumbers);	
	 
	      printReplicate(tr, numberOfTipsForInsertion, branches, thorough, replicates, inserts, insertNodes, insertNumbers);	  
	  
	  
	      printBothOpen("Classification Replicate [%d] completed within %f seconds\n", replicates, gettime() - rt);	

	      diversityStats(tr, numberOfTipsForInsertion, thorough, branches, insertNodes, perm, 
			     &queryDist, &queryNodes, &maxQueryDist, &maxQueryNodes, 
			     &treeDist,  &treeNodes,  &maxTreeDist,  &maxTreeNodes, replicates);
	    }
	  printBothOpen("\n\n");
	}
      else
	{	 
	  linkResults2Tree(tr, numberOfTipsForInsertion, insertNodes, branches, thorough, insertLikelihoods, insertNumbers);

	  diversityStats(tr, numberOfTipsForInsertion, thorough, branches, insertNodes, perm, 
			 &queryDist, &queryNodes, &maxQueryDist, &maxQueryNodes, 
			 &treeDist,  &treeNodes,  &maxTreeDist,  &maxTreeNodes, -1);
	}
    }
      
  printBothOpen("Overall Classification time: %f\n\n", gettime() - masterTime);			               	

  strcpy(labelledTreeFileName,         workdir);
  strcpy(originalLabelledTreeFileName, workdir);
  strcpy(classificationFileName,       workdir);
 
  
  strcat(labelledTreeFileName,         "RAxML_labelledTree.");
  strcat(originalLabelledTreeFileName, "RAxML_originalLabelledTree.");
  strcat(classificationFileName,       "RAxML_classification.");
  
  
  strcat(labelledTreeFileName,         run_id);
  strcat(originalLabelledTreeFileName, run_id);
  strcat(classificationFileName,       run_id);
 
  free(tr->tree_string);
  tr->treeStringLength *= 16;
  tr->tree_string  = (char*)malloc(tr->treeStringLength * sizeof(char));  
 
  treeFile = myfopen(labelledTreeFileName, "w");
  Tree2StringClassify(tr->tree_string, tr, numberOfTipsForInsertion, inserts, FALSE, -1, -1, (branchInfo*)NULL, -1, FALSE);
  fprintf(treeFile, "%s\n", tr->tree_string);    
  fclose(treeFile);
    
  treeFile = myfopen(originalLabelledTreeFileName, "w");
  Tree2StringClassify(tr->tree_string, tr, numberOfTipsForInsertion, inserts, FALSE, -1, -1, (branchInfo*)NULL, -1, TRUE);
  fprintf(treeFile, "%s\n", tr->tree_string);    
  fclose(treeFile);


  if(adef->rapidBoot)
    {
      root = findRoot(tr->start,  numberOfTipsForInsertion, tr->mxtips);           
      
      if(root == 0)
	root = findRoot(tr->start->back,  numberOfTipsForInsertion, tr->mxtips);
      
      if(root == 0)    
	printf("We are in deep shit, the tree can't be rooted\n");	      
      else
	{
	  char extendedTreeFileName[1024];
	  int cutoff, k;	     
	  
	  for(k = 20; k < 100; k += 5)
	    {
	      cutoff = (int)(((double)adef->multipleRuns * (double)k / 100.0) + 0.5);
	      sprintf(extendedTreeFileName, "%s.%d", labelledTreeFileName, k);	     
	      
	      treeFile = myfopen(extendedTreeFileName, "w");	  
	      Tree2StringClassify(tr->tree_string, tr, numberOfTipsForInsertion, inserts, TRUE, cutoff, numberOfBranches, tr->bInf, root, FALSE);	     
	      fprintf(treeFile, "%s\n", tr->tree_string);
	      fclose(treeFile);

	      printBothOpen("Least common ancestor file for cutoff %d (%d) written to file %s\n", cutoff, k, extendedTreeFileName);	      
	    }
	}
    }

  classificationFile = myfopen(classificationFileName, "w");
  
  for(i = 0; i < numberOfTipsForInsertion; i++)    
    for(j = 0; j < numberOfBranches; j++) 
      {       
	if(tr->bInf[j].countThem[i] > 0)	    
	  {
	    if(thorough)
	      fprintf(classificationFile, "%s I%d %d %8.20f\n", tr->nameList[inserts[i]], j, tr->bInf[j].countThem[i], tr->bInf[j].branches[i] / (double)(tr->bInf[j].countThem[i]));
	    else
	      fprintf(classificationFile, "%s I%d %d\n", tr->nameList[inserts[i]], j, tr->bInf[j].countThem[i]);
	  }
      }

  fclose(classificationFile);  

  printBothOpen("\n\nLabelled reference tree including branch labels and query sequences written to file: %s\n\n", labelledTreeFileName); 
  printBothOpen("Labelled reference tree with branch labels (without query sequences) written to file: %s\n\n", originalLabelledTreeFileName); 
  printBothOpen("Classification result file written to file: %s\n\n", classificationFileName);

  if(adef->rapidBoot)
    {
      double r = (double)(adef->multipleRuns);

      printBothOpen("Average Max Query Distance:  %f\n", maxQueryDist / r);
      printBothOpen("Average Max Tree Distance:   %f\n", treeDist / r);
      printBothOpen("Average Mean Query Distance: %f\n", queryDist / r);
      printBothOpen("Average Mean Tree Distance:  %f\n", treeDist / r);

      printBothOpen("Average Max Query  Count: %f\n", (double)maxQueryNodes / r);
      printBothOpen("Average Max Tree   Count: %f\n", (double)maxTreeNodes / r);
      printBothOpen("Average Mean Query Count: %f\n",   queryNodes / r);
      printBothOpen("Average Mean Tree  Count: %f\n\n", treeNodes / r);

      printBothOpen("Per Bootstrap replicate classification results written to files: RAxML_classificationReplicate.0.%s to\n", run_id);
      printBothOpen("                                                                 RAxML_classificationReplicate.%d.%s\n\n", (adef->multipleRuns - 1), run_id);
    }
  else
    {
      printBothOpen("Max Query Distance:  %f\n", maxQueryDist);
      printBothOpen("Max Tree Distance:   %f\n", maxTreeDist);
      printBothOpen("Mean Query Distance: %f\n", queryDist);
      printBothOpen("Mean Tree Distance:  %f\n", treeDist);

      printBothOpen("Max Query  Count: %d\n",   maxQueryNodes);
      printBothOpen("Max Tree   Count: %d\n",   maxTreeNodes);
      printBothOpen("Mean Query Count: %f\n",   queryNodes);
      printBothOpen("Mean Tree  Count: %f\n\n", treeNodes);
    }   
    
  exit(0);
}

void classifyML(tree *tr, analdef *adef)
{
  if(adef->dynamicAlignment == ALIGN_ML || adef->dynamicAlignment == ALIGN_MP)
    aligner(tr, adef);
  else
    classifier(tr, adef);
}
